Package: thematic
Title: Unified and Automatic 'Theming' of 'ggplot2', 'lattice', and
        'base' R Graphics
Version: 0.1.8
Authors@R: c(
    person("Carson", "Sievert", , "carson@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-4958-2844")),
    person("Barret", "Schloerke", , "barret@posit.co", role = "aut",
           comment = c(ORCID = "0000-0001-9986-114X")),
    person("Joe", "Cheng", , "joe@posit.co", role = "aut"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Theme 'ggplot2', 'lattice', and 'base' graphics based on a
    few choices, including foreground color, background color, accent
    color, and font family. Fonts that aren't available on the system, but
    are available via download on 'Google Fonts', can be automatically
    downloaded, cached, and registered for use with the 'showtext' and
    'ragg' packages.
License: MIT + file LICENSE
URL: https://rstudio.github.io/thematic/,
        https://github.com/rstudio/thematic
BugReports: https://github.com/rstudio/thematic/issues
Depends: R (>= 3.0.0)
Imports: farver, ggplot2 (>= 3.5.2), graphics, grDevices, grid,
        rappdirs, rlang, rstudioapi (>= 0.8), scales, utils
Suggests: bslib, Cairo, curl, gganimate, ggiraph, htmltools, jsonlite,
        knitr, lattice, ragg, rmarkdown, shiny (>= 1.5.0), showtext,
        stats, svglite, sysfonts, systemfonts, testthat, vdiffr (>=
        1.0.0), withr
Config/Needs/check: shinytest2, callr, sf, ggthemes, patchwork,
        gridExtra, tinytex, devtools, rversions
Config/Needs/routine: ggrepel, r-lib/evaluate
Config/Needs/website: GGally, RColorBrewer, patchwork,
        apreshill/quillt, r-lib/evaluate
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
Collate: 'auto.R' 'base.R' 'cache.R' 'gfonts.R' 'ggplot.R' 'globals.R'
        'hooks.R' 'knitr.R' 'lattice.R' 'onLoad.R' 'thematic-package.R'
        'thematic-save-plot.R' 'thematic.R' 'utils.R'
NeedsCompilation: no
Packaged: 2025-09-29 22:13:21 UTC; cpsievert
Author: Carson Sievert [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4958-2844>),
  Barret Schloerke [aut] (ORCID: <https://orcid.org/0000-0001-9986-114X>),
  Joe Cheng [aut],
  Posit Software, PBC [cph, fnd]
Maintainer: Carson Sievert <carson@posit.co>
Repository: CRAN
Date/Publication: 2025-09-29 23:40:03 UTC
Built: R 4.6.0; ; 2026-04-28 07:34:56 UTC; unix

# Generated by roxygen2: do not edit by hand

export(auto_config)
export(auto_config_get)
export(auto_config_set)
export(auto_resolve_theme)
export(default_device)
export(font_cache_set)
export(font_spec)
export(okabe_ito)
export(sequential_gradient)
export(thematic_get_mixture)
export(thematic_get_option)
export(thematic_get_theme)
export(thematic_local_theme)
export(thematic_off)
export(thematic_on)
export(thematic_rmd)
export(thematic_save_plot)
export(thematic_set_theme)
export(thematic_shiny)
export(thematic_theme)
export(thematic_with_theme)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.cur)
importFrom(grDevices,dev.new)
importFrom(grDevices,dev.off)
importFrom(grDevices,dev.set)
importFrom(grDevices,devAskNewPage)
importFrom(grDevices,palette)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,plot.new)
importFrom(grid,gpar)
importFrom(grid,grid.draw)
importFrom(grid,grid.newpage)
importFrom(grid,grid.text)
importFrom(grid,pushViewport)
importFrom(grid,seekViewport)
importFrom(grid,upViewport)
importFrom(rlang,"%|%")
importFrom(rlang,is_installed)
importFrom(rlang,list2)
importFrom(rstudioapi,getThemeInfo)
importFrom(rstudioapi,readRStudioPreference)
importFrom(rstudioapi,versionInfo)
importFrom(utils,assignInNamespace)
importFrom(utils,capture.output)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,getS3method)
importFrom(utils,modifyList)
importFrom(utils,packageVersion)
importFrom(utils,unzip)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# thematic <a href='https://rstudio.github.io/thematic/'><img src='man/figures/logo.png' align="right" height="138.5" style="margin:10px;" /></a>

<!-- badges: start -->

[![R build
status](https://github.com/rstudio/thematic/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/thematic)
[![CRAN
status](https://www.r-pkg.org/badges/version/thematic)](https://CRAN.R-project.org/package=thematic)
<!-- badges: end -->

Simplified theming of `{ggplot2}`, `{lattice}`, and `{base}` R graphics.
In addition to providing a [centralized
approach](https://rstudio.github.io/thematic/articles/custom.html) to
styling R graphics, `{thematic}` also enables [automatic
styling](https://rstudio.github.io/thematic/articles/auto.html) of R
plots in Shiny, R Markdown, and RStudio.

## Installation

Install the stable release of `{thematic}` on
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("thematic")
```

[Auto theming in
Shiny](https://rstudio.github.io/thematic/articles/auto.html#shiny)
requires `{shiny}` 1.5.0 or higher:

``` r
install.packages("shiny")
```

[Auto theming in R
Markdown](https://rstudio.github.io/thematic/articles/auto.html#rmd)
requires `{rmarkdown}` 2.7 or higher:

``` r
install.packages("rmarkdown")
```

Using themes with [custom
fonts](https://rstudio.github.io/thematic/articles/fonts.html) works
best if you have `{showtext}` and/or `{ragg}` installed.

``` r
install.packages("showtext")
install.packages("ragg")
```

## Overview

`{thematic}`’s [auto
theming](https://rstudio.github.io/thematic/articles/auto.html) gives R
plots the ability to style themselves inside [Shiny](#shiny) (via CSS),
[RStudio](#rstudio) (via [RStudio
themes](https://support.posit.co/hc/en-us/articles/115011846747-Using-Themes-in-the-RStudio-IDE)),
and [R Markdown](#rmarkdown) (via
[`{bslib}`](https://rstudio.github.io/bslib/)).

### Shiny

Call `thematic_shiny()` before launching a Shiny app to enable
`{thematic}` for every `plotOutput()` inside the app. If no values are
provided to `thematic_shiny()`, each `plotOutput()` uses the app’s CSS
colors to inform new R plotting defaults. If the app uses [Google
Fonts](https://fonts.google.com/) (and you have `{showtext}` and/or
`{ragg}` installed), you may safely provide `font = "auto"` to
`thematic_shiny()`, which also translates CSS fonts to R. Here’s an
example with the [Pacifico](https://fonts.google.com/specimen/Pacifico)
font:

``` r
library(shiny)
library(ggplot2)
library(thematic)
# In order for auto/custom fonts to work properly, you'll want
# either the ragg (or showtext) package installed
library(ragg)

# If you want `{ragg}` to handle the font rendering in a Shiny app
options(shiny.useragg = TRUE)

# Call thematic_shiny() prior to launching the app, to change
# R plot theming defaults for all the plots generated in the app
thematic_shiny(font = "auto")

ui <- fluidPage(
  # bslib makes it easy to customize CSS styles for things
  # rendered by the browser, like tabsetPanel()
  # https://rstudio.github.io/bslib
  theme = bslib::bs_theme(
    bg = "#002B36", fg = "#EEE8D5", primary = "#2AA198",
    # bslib also makes it easy to import CSS fonts
    base_font = bslib::font_google("Pacifico")
  ),
  tabsetPanel(
    type = "pills",
    tabPanel("ggplot", plotOutput("ggplot")),
    tabPanel("lattice", plotOutput("lattice")),
    tabPanel("base", plotOutput("base"))
  )
)

server <- function(input, output) {
  output$ggplot <- renderPlot({
    ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars), color = factor(cyl))) +
      geom_point() +
      ggrepel::geom_text_repel()
  })
  output$lattice <- renderPlot({
    lattice::show.settings()
  })
  output$base <- renderPlot({
    image(volcano, col = thematic_get_option("sequential"))
  })
}

shinyApp(ui, server)
```

<img src="man/figures/auto-before.png" width="90%" style="display: block; margin: auto;" />

<img src="man/figures/auto-after.png" width="90%" style="display: block; margin: auto;" />

### RStudio

Call `thematic_on()` before generating plots inside RStudio to have all
subsequent plots shown in the “Plots” viewing pane to reflect your
RStudio theme. Note that `thematic_on()` enables `{thematic}` for the
remainder of the R session, but you can use `thematic_off()` to disable
([or `thematic_theme()` for one-off use of
`{thematic}`](https://rstudio.github.io/thematic/articles/scope.html#one-time-use)).
Here’s an example of how `{thematic}` can intelligently adapt each plot
to the current RStudio theme:

<img src="https://i.imgur.com/Bvbdn2B.gif" width="100%" style="display: block; margin: auto;" />

### R Markdown

Call `thematic_rmd()` before generating plots inside R Markdown to have
all subsequent plots within the document reflect the relevant theme. In
a static (i.e., non-`runtime: shiny`) R Markdown context, auto-theming
only works with [`{bslib}`-powered
`rmarkdown::html_document()`](https://rstudio.github.io/bslib/#r-markdown-usage)
(as in the example below), but in other situations you may also [provide
colors and fonts](#custom) explicitly to `thematic_rmd()`.

<img src="vignettes/html-document.svg" width="100%" style="display: block; margin: auto;" />

### Custom theming

By default, `{thematic}` attempts to detect the relevant background,
foreground, and accent colors. However, you may also specify these
settings more directly by providing relevant color and fonts directly to
`thematic_on()` (or `thematic_shiny()`/`thematic_rmd()`).

``` r
library(ggplot2)
thematic::thematic_on(bg = "#222222", fg = "white", accent = "#0CE3AC", font = "Oxanium")

ggp <- ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars), color = factor(cyl))) +
  geom_point() +
  ggrepel::geom_text_repel()
ggp
```

<img src="man/figures/README-ggrepel-1.png" width="80%" style="display: block; margin: auto;" />

`{thematic}` works by setting new global defaults that can always be
overridden with plot-specific `theme()`-ing code:

``` r
ggp + theme(text = element_text(colour = "purple"))
```

<img src="man/figures/README-ggrepel2-1.png" width="80%" style="display: block; margin: auto;" />

To use a “complete” `{ggplot2}` theme with `{thematic}` (e.g.,
`theme_bw()`, `theme_minimal()`, etc), use `theme_set()` to set the
theme globally. This way `{thematic}` has the opportunity to preserve
the complete theme’s styling semantics when changing global defaults
(e.g., `theme_bw()` uses the same fill color for the panel and plot
background, which is semantically different from the `theme_gray()`
default):

``` r
theme_set(theme_bw())
ggp
```

<img src="man/figures/README-ggrepel3-1.png" width="80%" style="display: block; margin: auto;" />

In addition to setting new defaults for main colors and fonts,
`{thematic}` also sets defaults for `qualitative` (and `sequential`)
colorscales. See the [custom themes
article](https://rstudio.github.io/thematic/articles/custom.html) to
learn more about how to customize those defaults.

## Learn more

- See the [auto theming
  article](https://rstudio.github.io/thematic/articles/auto.html) to
  gain a more detailed understanding of how auto theming make styling R
  plots easier in Shiny, R Markdown, and RStudio.
- See the [custom themes
  article](https://rstudio.github.io/thematic/articles/custom.html) for
  more on `{thematic}`’s theming options as well as how they interact
  with `{ggplot2}`, `{lattice}`, and `{base}`.
- See the [fonts
  article](https://rstudio.github.io/thematic/articles/fonts.html) for
  more on using Google Fonts with `{thematic}`.
- See the [scoping
  article](https://rstudio.github.io/thematic/articles/scope.html) for
  more about scoping `{thematic}` to individual plots.

## Run some examples

Below is a link to an **RStudio Cloud** instance with some ready to run
`{thematic}` examples:

<div>

<a href="https://rstudio.cloud/project/1208127" target="_blank">
<img src="man/figures/thematic-test-drive.svg" alt="RStudio Cloud Example" height="80px" style="display: block; margin: 0 auto;">
</a>

</div>

## Code of Conduct

`{thematic}` is released with a [Contributor Code of
Conduct](https://contributor-covenant.org/version/2/0/CODE_OF_CONDUCT.html).
By
[contributing](https://github.com/rstudio/thematic/blob/main/CONTRIBUTING.md)
to this project, you agree to abide by its terms.

library(shiny)
library(ggplot2)
library(ggthemes)
library(thematic)
library(dplyr)
library(showtext)
library(Cairo)
library(bslib)


ui <- fluidPage(
  theme = bs_theme(bootswatch = "darkly"),
  selectInput(
    "theme", "Choose a theme",
    choices = grep("^theme_", getNamespaceExports("ggthemes"), value = TRUE)
  ),
  fluidRow(
    column(6, tags$h4("Before thematic"), plotOutput("theme_off")),
    column(6, tags$h4("After thematic"), plotOutput("theme_on"))
  ),
  verbatimTextOutput("cache")
)

p <- ggplot(iris, aes(x = Sepal.Length, y = Sepal.Width, color =  Species)) +
  geom_point()

server <- function(input, output, session) {

  observe({
    theme_fun <- match.fun(input$theme)
    theme_set(theme_fun())
  }, priority = 2)

  output$theme_on <- renderPlot({
    input$theme
    thematic_with_theme(thematic_theme(font = "Pacifico"), print(p))
  })

  output$theme_off <- renderPlot({input$theme; print(p)})

}

shinyApp(ui, server)

# view all shinytest differences inside package tests
# (only for internal use)
# https://github.com/rstudio/thematic/blob/master/CONTRIBUTING.md
view_shinytest_diffs <- function(path = ".", suffix = "auto", ...) {
  if (identical(suffix, "auto")) {
    branch <- system("git rev-parse --abbrev-ref HEAD", intern = TRUE)
    pieces <- strsplit(branch, "-")[[1]]
    platform <- pieces[[length(pieces)]]
    suffix <- switch(
      platform,
      Windows = "win",
      macOS = "mac",
      Linux = "linux",
      ""
    )
  }
  test_path <- file.path(path, "tests", "testthat")
  test_path <- normalizePath(test_path, mustWork = TRUE)
  test_dirs <- grep("tests/shinytest/.*-current", list.dirs(test_path), value = TRUE)
  app_dirs <- dirname(dirname(dirname(test_dirs)))
  for (dir in app_dirs) {
    getFromNamespace("viewTestDiff", "shinytest")(appDir = dir, suffix = suffix, interactive = TRUE, ...)
  }
}

library(shiny)
library(ggplot2)
library(thematic)

thematic_shiny("black", "white", font = font_spec("Pacifico", 1.25, update = TRUE))

ui <- fluidPage(
  imageOutput("ragg_png")
)

render_image <- function(expr) {
  snapshotPreprocessOutput(renderImage(expr), function(value) {})
}

server <- function(input, output, session) {
  output$ragg_png <- render_image({
    file <- thematic_save_plot(
      qplot(1:10) + ggtitle("ragg::agg_png"),
      device = ragg::agg_png,
      width = 800, height = 400
    )
    list(src = file, width = 800, height = 400)
  })
}

shinyApp(ui, server)

.onLoad <- function(libname, pkgname) {
  # Leverage rsconnect's pre-deployment hook to copy the font
  # cache over to a directory that we can use on the server
  # https://github.com/rstudio/rsconnect/pull/295/files
  pre_deploy <- getOption("rsconnect.pre.deploy")
  options(
    rsconnect.pre.deploy = function(app_dir) {
      if (is.function(pre_deploy)) {
        pre_deploy(app_dir)
      }
      font_cache_set(app_dir)
    }
  )

  # CRAN versions
  register_upgrade_message("shiny", "1.5.0")   # shiny::getCurrentOutputInfo()
  register_upgrade_message("ggplot2", "3.3.2") # proper scale defaults
  register_upgrade_message("ragg", "0.2")      # important bug fixes
  register_upgrade_message("rmarkdown", "2.7") # bslib integration

  install_knitr_hooks()
}


# Essentially verbatim from shiny:::register_upgrade_message
register_upgrade_message <- function(pkg, version, location = NULL) {

  msg <- sprintf(
    "This version of thematic is designed to work with %s version %s or higher. ",
    pkg, version
  )

  if (length(location)) {
    msg <- paste0(msg, sprintf("Consider upgrading via remotes::install_github('%s')", location))
  }

  if (pkg %in% loadedNamespaces() && !is_available(pkg, version)) {
    packageStartupMessage(msg)
  }

  # Always register hook in case pkg is loaded at some
  # point the future (or, potentially, but less commonly,
  # unloaded & reloaded)
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      if (!is_available(pkg, version)) packageStartupMessage(msg)
    }
  )
}
is_available <- function(package, version = NULL) {
  installed <- nzchar(system.file(package = package))
  if (is.null(version)) {
    return(installed)
  }
  installed && isTRUE(utils::packageVersion(package) >= version)
}

library(shiny)
library(thematic)

theme_base <- thematic_theme("black", "white", accent = "purple")
theme_ggplot <- thematic_theme("black", "white", accent = "red")
theme_lattice <- thematic_theme("black", "white", accent = "orange")

ui <- fluidPage(
  plotOutput("base"),
  plotOutput("ggplot"),
  plotOutput("lattice")
)

server <- function(input, output, session) {

  output$base <- renderPlot({
    thematic_with_theme(theme_base, {
      plot(1:10, 1:10, col = thematic_get_option("accent"))
    })
  })

  output$ggplot <- renderPlot({
    thematic_with_theme(theme_ggplot, {
      ggplot2::qplot(1:10, 1:10, color = 1:10)
    })
  })

  output$lattice <- renderPlot({
    thematic_with_theme(theme_lattice, {
      lattice::show.settings()
    })
  })

}

shinyApp(ui, server)

library(shiny)
library(ggplot2)
library(thematic)

# The device we're testing
device <- "CairoPNG"
ext <- "png"

ui <- fluidPage(
  imageOutput("ggplot"),
  imageOutput("lattice"),
  imageOutput("base")
)

render_image <- function(expr) {
  snapshotPreprocessOutput(renderImage(expr, deleteFile = FALSE), function(value) {})
}

image_info <- function(type, ext) {
  list(
    src = paste0(type, "-1.", ext),
    width = 600, height = 400
  )
}

server <- function(input, output, session) {
  # Knit all the Rmds
  # Note that this is done inside the server function
  # to avoid a timeout issue with shinytest::testApp()
  rmd_txt <- knitr::knit_expand("../template_device.Rmd", device = sprintf("'%s'", device))
  res <- callr::r(function(...) { knitr::knit2html(...) }, args = list(text = rmd_txt))

  output$ggplot <- render_image(image_info("ggplot", ext))
  output$lattice <- render_image(image_info("lattice", ext))
  output$base <- render_image(image_info("base", ext))

  onFlush(function() {
    unlink(dir(pattern = paste0("\\.", ext)))
  })
}

shinyApp(ui, server)

base_palette_set <- function(theme = .globals$theme) {
  base_palette_restore()
  codes <- theme$qualitative
  .globals$base_palette <- if (isTRUE(is.na(codes))) {
    attempt_palette()
  } else {
    attempt_palette(codes)
  }
}

base_palette_restore <- function() {
  if (is.null(.globals$base_palette)) return()
  attempt_palette(.globals$base_palette)
  rm("base_palette", envir = .globals)
}

base_params_set <- function(theme = .globals$theme) {
  base_params_restore()
  params <- list()
  bg <- theme$bg
  if (length(bg)) {
    params <- c(params, attempt_par(bg = bg))
  }
  fg <- theme$fg
  if (length(fg)) {
    params <- c(params, attempt_par(
      fg = fg,
      col.axis = fg,
      col.lab = fg,
      col.main = fg,
      col.sub = fg
    ))
  }
  font <- theme$font
  if (length(font$family)) {
    params <- c(params, attempt_par(
      family = font$family,
      cex.axis = font$scale,
      cex.lab = font$scale,
      cex.main = font$scale * 1.2,
      cex.sub = font$scale
    ))
  }

  .globals$base_params <- params
}

base_params_restore <- function() {
  if (is.null(.globals$base_params)) return()
  do.call(attempt_par, .globals$base_params)
  rm("base_params", envir = .globals)
}

attempt_par <- function(...) {
  attempt_(par(...))
}

attempt_palette <- function(...) {
  attempt_(palette(...))
}

attempt_ <- function(expr) {
  if (is_null_device()) {
    attempt_with_new_device(expr)
  } else {
    force(expr)
  }
}

#' Configure auto theming behavior
#'
#' Auto theming is really only "guaranteed" to work inside of a **shiny**
#' runtime. In any other context, auto theming is based on a set of heuristics,
#' which won't fit every use case. As a workaround, this function allows one
#' to configure both a preference for specific auto values (e.g., `bg`, `fg`, etc)
#' as well as the `priority` that certain information should receive.
#'
#' @details Configuring auto theming behavior is especially useful
#' for developers of a custom rmarkdown output document that wish to
#' have more sensible auto theming behavior for users of the document.
#' In particular, by having the output document call `auto_config_set()`
#' "pre-knit" with the document's styling preferences (and restoring the
#' old defaults "post-knit"), users of the output document can then simply
#' call `thematic_on()` within their document to use those preferences.
#'
#' @details Call this function with no arguments to get the current auto defaults.
#'
#' @inheritParams thematic_on
#' @param priority the order of priority to use when resolving auto values.
#' Possible values include:
#'   * `"shiny"`: use [shiny::getCurrentOutputInfo()] values (if any) to resolve auto values.
#'   * `"config"`: use the values provided to this function (if any) to resolve auto values.
#'   * `"bslib"`: use `bslib::bs_get_variables()` values (if any)
#'     to resolve auto values (only relevant when knitr is in progress).
#'   * `"rstudio"`: use [rstudioapi::getThemeInfo()] values (if any) to resolve auto values.
#'
#' @return a config (list-like) object.
#'
#' @rdname auto-config
#' @export
#' @examples
#' old_config <- auto_config_set(auto_config("black", "white"))
#' thematic_with_theme(
#'   thematic_theme(), {
#'     plot(1:10, 1:10)
#'  })
#' auto_config_set(old_config)
auto_config <- function(bg = NULL, fg = NULL, accent = NULL, font = NULL,
                        priority = c("shiny", "config", "bslib", "rstudio")) {
  cols <- dropNulls(list(bg = bg, fg = fg, accent = accent))
  config <- lapply(cols, function(x) {
    if (isTRUE(is.na(x))) x else parse_any_color(x)
  })
  if (!is.null(font)) {
    config$font <- as_font_spec(font)
  }
  priority <- priority %||% priorities()
  config$priority <- match.arg(priority, priorities(), several.ok = TRUE)
  structure(config, class = "thematic_auto_config")
}

#' @rdname auto-config
#' @param config a `auto_config()` object.
#' @export
auto_config_set <- function(config) {
  if (!inherits(config, "thematic_auto_config")) {
    stop("`config` must be a `auto_config()` object", call. = FALSE)
  }
  old_config <- auto_config_get()
  .globals$auto_config <- config
  invisible(old_config)
}

#' @rdname auto-config
#' @export
auto_config_get <- function() {
  # This should contain priority even if a config hasn't been set
  config <- utils::modifyList(auto_config(), .globals$auto_config %||% list())
  structure(config, class = "thematic_auto_config")
}

priorities <- function() {
  c("shiny", "config", "bslib", "rstudio")
}

#' Resolve auto values
#'
#' Resolves `'auto'` values based on the current execution environment
#' and configuration (i.e., [auto_config_get()]).
#'
#' @return The `theme` object with resolved `'auto'` values.
#'
#' @param theme a `thematic_theme()` object.
#' @export
#' @seealso [auto_config_set()]
#' @examples
#'
#' old_config <- auto_config_set(auto_config(bg = "black", fg = "white"))
#'
#' # Resolving auto values in local theme objects
#' theme <- thematic_theme()
#' theme[c("bg", "fg")]
#' theme <- auto_resolve_theme(theme)
#' theme[c("bg", "fg")]
#'
#' # By default, auto values are resolved when accessing
#' # global theme options
#' thematic_on()
#' thematic_get_option("bg", resolve = FALSE)
#' thematic_get_option("bg")
#' thematic_off()
#'
#' auto_config_set(old_config)
#'
auto_resolve_theme <- function(theme) {
  if (length(theme) == 0) {
    return(theme)
  }
  if (!is_thematic_theme(theme)) {
    stop("`theme` must be a `thematic_theme()` object", call. = FALSE)
  }

  auto_color_info <- lapply(auto_config_get()$priority, function(x) {
    switch(
      x,
      shiny = shiny_output_info(),
      config = auto_config_get(),
      bslib = bs_theme_colors(),
      rstudio = rs_theme_colors(),
      stop("`priority` of '", x, "' is not implemented", call. = FALSE)
    )
  })

  # Resolve auto colors, if relevant
  for (col in c("bg", "fg", "accent")) {
    if (!is_auto(theme[[col]])) {
      next
    }
    theme[[col]] <- Reduce(`%||%`, lapply(auto_color_info, `[[`, col)) %||% theme[[col]]
    if (isTRUE("auto" == theme[[col]])) {
      maybe_warn(
        "thematic was unable to resolve `", col, "='auto'`. ",
        "Try providing an actual color (or `NA`) to the `", col, "` argument of `thematic_on()`. ",
        "By the way, 'auto' is only officially supported in `shiny::renderPlot()`, ",
        "some rmarkdown scenarios (specifically, `html_document()` with `theme!=NULL`), ",
        "in RStudio, or if `auto_config_set()` is used.",
        id = paste0("auto-detection-failure-", col)
      )
      theme[[col]] <- switch(col, bg = "white", fg = "black", NA)
    } else {
      theme[[col]] <- parse_any_color(theme[[col]])
    }

    # Retain auto class so that the _next_ time this hook
    # gets called we know to resolve the value again
    # (e.g., renderPlot() executes again, but this time with different styling)
    theme[[col]] <- as_auto(theme[[col]])
  }

  # resolve sequential, if necessary
  sequential_func <- attr(theme$sequential, "sequential_func")
  if (is.function(sequential_func)) {
    theme$sequential <- do.call(sequential_func, theme[c("bg", "fg", "accent")])
  }

  # Make sure we can parse any non-missing colors
  for (col in c("bg", "fg", "accent", "qualitative", "sequential")) {
    if (isTRUE(is.na(theme[[col]]))) next
    val <- vapply(theme[[col]], parse_any_color, character(1), USE.NAMES = FALSE)
    # Retain auto class (see comment above)
    theme[[col]] <- if (is_auto(theme[[col]])) as_auto(val) else val
  }

  # Retain the function that created sequential codes, if necessary
  if (is.function(sequential_func)) {
    theme$sequential <- structure(theme$sequential, sequential_func = sequential_func)
  }

  # Resolve fonts
  if (any(vapply(theme$font, is_auto, logical(1)))) {
    # Note the similarity to resolution of auto colors
    auto_font_info <- lapply(auto_config_get()$priority, function(x) {
      switch(
        x,
        shiny = shiny_font_spec(shiny_output_info()$font),
        config = auto_config_get()$font,
        bslib = bs_font_spec(),
        rstudio = rs_font_spec(),
        stop("`priority` of '", x, "' is not implemented", call. = FALSE)
      )
    })
    spec <- Reduce(`%||%`, auto_font_info) %||% font_spec()

    # As with colors above, make sure to retain the auto class so that the
    # _next_ time this hook gets called we know to resolve the value again
    for (key in names(spec)) {
      if (is_auto(theme$font[[key]])) {
        theme$font[[key]] <- as_auto(spec[[key]])
      }
    }

  } else {
    theme$font <- as_font_spec(theme$font)
  }

  theme
}

# ------------------------------------------------------------
# Colors
# ------------------------------------------------------------

shiny_output_info <- function() {
  if (!"shiny" %in% loadedNamespaces()) return(NULL)
  info <- tryNULL(shiny::getCurrentOutputInfo())
  # Return early if we're not in any output context
  if (is.null(info)) return(NULL)
  # Return early with a message to update shiny if the relevant
  # info isn't populated
  nms <- c("bg", "fg", "accent", "font")
  missing <- setdiff(nms, names(info))
  # In this case, we're in an output context, but `.shiny-report-theme` info is missing
  if (length(missing)) {
    if (is_available("shiny", "1.5.0")) {
      maybe_warn(
        "Auto-theming with shiny requires an output context that has reports it's CSS ",
        "styles. Do you want to create this plot inside `renderPlot()` (or another output ",
        "container with a .shiny-report-theme class)?",
        id = "missing-css-info"
      )
    } else {
      maybe_warn(
        "Auto-theming with shiny requires v1.5.0 or higher",
        id = "upgrade-shiny"
      )
    }
    return(NULL)
  }
  res <- lapply(info[nms], function(x) { x() })
  rlang::set_names(res, nms)
}

bs_theme_colors <- function(theme = bslib::bs_global_get()) {
  if (!in_html_document()) return(NULL)

  cols <- if ("3" %in% theme_version(theme)) {
    bs_theme_vars(theme, c("body-bg", "text-color", "link-color"))
  } else {
    bs_theme_vars(theme, c("body-bg", "body-color", "link-color"))
  }

  rlang::set_names(cols, c("bg", "fg", "accent"))
}

bs_theme_vars <- function(...) {
  utils::getFromNamespace("bs_get_variables", "bslib")(...)
}

theme_version <- function(...) {
  utils::getFromNamespace("theme_version", "bslib")(...)
}


rs_theme_colors <- function() {
  # Try to get theme info from rstudioapi::getThemeInfo().
  # Hopefully someday this'll return font/accent info
  # https://github.com/rstudio/rstudioapi/issues/174
  info <- try_get_theme_info()
  if (is.null(info)) return(NULL)

  # These colors were taken manually from the theme preview
  # (they are the token color)
  info$accent <- switch(
    info$editor,
    Ambiance = "#CFB171",
    Chaos = "#27759B",
    Chrome = "#98108D",
    Clouds = "#98108D",
    `Clouds Midnight` = "#9A8767",
    Cobalt = "#F7A600",
    `Crimson Editor` = "#2918FF",
    Dawn = "#80523F",
    Dracula = "#F785CE",
    Dreamweaver = "#2918FF",
    Eclipse = "#86058D",
    `Idle Fingers` = "#CB8334",
    Katzenmilch = "#7255B3",
    `Kr Theme` = "#9FA695",
    Material = "#DC9D67",
    Merbivore = "#F37900",
    `Merbivore Soft` = "#F48A3B",
    `Mono Industrial` = "#ABA86D",
    Monokai = "#F0297D",
    `Pastel On Dark` = "#8486DF",
    `Solarized Dark` = "#90A300",
    `Solarized Light` = "#90A300",
    Textmate = "#2918FF",
    Tomorrow = "#9265B3",
    `Tomorrow Night` = "#B99FC4",
    `Tomorrow Night 80s` = "#CFA4D3",
    `Tomorrow Night Blue` = "#EBC3FF",
    `Tomorrow Night Bright` = "#C8A2DE",
    Twilight = "#CFB171",
    `Vibrant Ink` = "#F57000",
    Xcode = "#C706AF",
    NA
  )
  rlang::set_names(
    info[c("background", "foreground", "accent")],
    c("bg", "fg", "accent")
  )
}


# ------------------------------------------------------------
# Fonts
# ------------------------------------------------------------

shiny_font_spec <- function(font) {
  if (!length(font)) return(NULL)
  font_spec(
    font$families,
    scale = size_to_scale(font$size)
  )
}

bs_font_spec <- function(theme = bslib::bs_global_get()) {
  if (!in_html_document()) return(NULL)

  family <- bs_theme_vars(theme, "font-family-base")
  families <- strsplit(gsub('"', '', family), ", ")[[1]]
  size <- bs_theme_vars(theme, "font-size-base")
  font_spec(families, scale = size_to_scale(size))
}

in_html_document <- function() {
  if (!getOption("knitr.in.progress", FALSE)) return(FALSE)
  if (!is_installed("bslib")) return(FALSE)
  !is.null(utils::getFromNamespace("bs_global_get", "bslib")())
}

# Translate CSS font-size to font_spec(scale = ...)
# https://developer.mozilla.org/en-US/docs/Web/CSS/length
size_to_scale <- function(size, pointsize = 12) {
  if (length(size) != 1) {
    stop("Expect font size to be of length 1.", call. = FALSE)
  }

  size <- sub("^\\s+", "", sub("\\s+$", "", size))

  # Based on https://stackoverflow.com/a/5912657/1583084
  size <- switch(
    size,
    `xx-small` = "50%",
    `x-small`  = "62.5%",
    small      = ,
    smaller    = "80%",
    medium     = "100%",
    large      = ,
    larger     = "112.5%",
    `x-large`  = "150%",
    `xx-large` = "200%",
    size
  )

  # Translate some important relative units (1rem is the default for BS4).
  # Yes, these will be wrong if the reference size is different from
  # the default of 1(r)em = 16px = 12pt...shrug
  if (grepl("[0-9]+r?em$", size)) {
    return(as.numeric(sub("r?em$", "", size)))
  }
  if (grepl("[0-9]+%", size)) {
    return(as.numeric(sub("%$", "", size)) / 100)
  }

  # ------------------------------------------------------
  # Translate all absolute units to inches
  # https://developer.mozilla.org/en-US/docs/Web/CSS/length
  # ------------------------------------------------------
  from_to <- function(x, from, to, factor) {
    pattern <- paste0("[0-9]+", from, "$")
    if (!grepl(from, x)) return(x)
    paste0(as.numeric(sub(from, "", x)) * factor, to)
  }

  size <- from_to(size, "px", "in", 1/96)
  size <- from_to(size, "cm", "in", 2.54)
  size <- from_to(size, "mm", "in", 25.4)
  size <- from_to(size, "pc", "in", 1/6)
  pts <- from_to(size, "in", "pt", 72)

  tryCatch(
    as.numeric(sub("pt", "", pts)) / (.globals$device$args$pointsize %||% 12),
    warning = function() {
      maybe_warn(
        "CSS font-size unit of '", size, "' not supported by thematic.",
        id = "unsupported-font-size-unit"
      )
      1
    }
  )
}


rs_font_spec <- function() {
  if (!is_rstudio()) return(NULL)
  if (!in_rstudio_gd()) return(NULL)

  # readRStudioPreference was introduced in RStudio 1.3
  pts <- tryCatch(readRStudioPreference("font_size_points"), error = function(e) 12)

  # Note that server_editor_font appears in RStudio 1.4
  # so it'll take awhile for this to be widely supported
  family <- tryCatch(readRStudioPreference("server_editor_font"), error = function(e) "")

  if (isTRUE(nzchar(family))) {
    # TODO: respect global device pointsize
    return(font_spec(family, scale = pts / 12))
  }

  if (identical(versionInfo()$mode, "server")) {
    message("Auto font detection in RStudio Server requires RStudio 1.4 or higher.")
    return(NULL)
  }

  # Try and read RStudio Desktop editor font preference
  # https://support.rstudio.com/hc/en-us/articles/200534577-Resetting-RStudio-Desktop-s-State
  family <- grep(
    '\\s*"?font.*fixedWidth\\s*=',
    tryNULL(rstudio_desktop_prefs()),
    value = TRUE
  )
  # TODO: if font hasn't been specified, this font.fixedWidth wont appear...
  # is it correct to just use the default font?
  if (!length(family)) {
    return(NULL)
  }
  family <- strsplit(family, "=")[[1]][2]
  family <- sub('^\\s*"?', '', sub('"?\\s*;?\\s*$', '', family))
  # TODO: respect global device pointsize
  font_spec(family, scale = pts / 12)
}

rstudio_desktop_prefs <- function() {
  if (.Platform$OS.type == "windows") {
    return(readLines(
      file.path(Sys.getenv("APPDATA"), "RStudio", "desktop.ini")
    ))
  }
  sys <- Sys.info()[["sysname"]]
  if (sys == "Darwin") {
    return(
      system("defaults read com.rstudio.desktop", intern = TRUE)
    )
  }
  readLines("~/.config/RStudio/desktop.ini")
}


tag_auto <- function(x) {
  if (identical(x, "auto")) as_auto(x) else x
}

as_auto <- function(x) {
  if (is_auto(x)) return(x)
  oldClass(x) <- c("thematic_auto", oldClass(x))
  x
}

is_auto <- function(x) {
  inherits(x, "thematic_auto")
}

test_that("Can influence auto resolution with config", {
  # TODO: add test(s) with different priority settings
  old_config <- auto_config_set(auto_config("black", "white", NA, font = "Pacifico"))
  on.exit(auto_config_set(old_config), add = TRUE)

  theme <- thematic_theme(font = "auto")
  expect_true(theme$bg == "auto")
  expect_true(theme$fg == "auto")
  expect_true(theme$accent == "auto")
  theme <- auto_resolve_theme(theme)
  expect_true(theme$bg == "black")
  expect_true(theme$fg == "white")
  expect_true(is.na(theme$accent))
  expect_true(theme$font$families == "Pacifico")
})

test_that("priorities() values match priority arg", {
  expect_equal(priorities(), eval(formals(auto_config)$priority))
})

test_that("thematic_theme() default values are consistent", {
  expect_equal(formals(thematic_on), formals(thematic_theme))
  expect_equal(formals(thematic_on), formals(thematic_rmd))
  fmls <- formals(thematic_shiny)
  fmls$session <- NULL
  expect_equal(formals(thematic_on), fmls)
})

library(shiny)
library(ggplot2)
library(thematic)

thematic_shiny("black", "white", font = font_spec("Pacifico", 1.25, update = TRUE))

ui <- fluidPage(
  imageOutput("svg")
)

render_image <- function(expr) {
  snapshotPreprocessOutput(renderImage(expr), function(value) {})
}

server <- function(input, output, session) {
  output$svg <- render_image({
    file <- thematic_save_plot(
      qplot(1:10) + ggtitle("grDevices::svg"),
      device = grDevices::svg, filename = tempfile(fileext = ".svg"),
      antialias = "none",
      width = 8, height = 4
    )
    list(src = file, width = 800, height = 400)
  })
}

shinyApp(ui, server)

# vdiffr:::print_plot.ggplot adds ggplot2::theme_test(), which destroys our theming defaults
registerS3method(
  "print_plot", "ggplot", function(p, title = "") { print(p) },
  envir = asNamespace("vdiffr")
)

expect_doppelganger <- function(name, p, ...) {
  # Visual testing is currently only done locally (via devtools::test()).
  # This is because we want to .Rbuildignore the _snaps folder, which means
  # they won't be available when R CMD check is run on CI.
  # Suppose we could add another step to the CI workflow to run devtools::test(),
  # but I'm going to wait on that until we agree on a standard (for shiny-workflows).
  testthat::skip_on_ci()
  vdiffr::expect_doppelganger(name, p, ...)
}

# some plots have randomized output
set.seed(101)

#' @importFrom graphics par plot plot.new
#' @importFrom grid grid.newpage grid.draw grid.text gpar seekViewport pushViewport upViewport
#' @importFrom grDevices col2rgb dev.new dev.cur dev.off dev.set devAskNewPage palette colorRampPalette
#' @importFrom utils capture.output download.file unzip packageVersion modifyList getFromNamespace getS3method assignInNamespace
#' @importFrom rlang is_installed list2 %|%
#' @importFrom rstudioapi getThemeInfo versionInfo readRStudioPreference
NULL

#' Control the directory used for font caching
#'
#' The default directory used for font caching is system dependent;
#' and thus, not very portable from machine to machine. Use this
#' function to move thematic's cache to a new path. This is primarily
#' useful for making font cache relative to a shiny app directory,
#' so that, when the app is deployed, the cache deploys with it.
#'
#' @param path a filepath for the new cachine directory.
#' @param cleanup whether or not to remove font files from the
#' previously used caching directory (after copying to the new location).
#'
#' @return Returns the previously used caching directory.
#' @seealso [thematic_on()], [font_spec()]
#'
#' @export
#' @examples
#' \dontrun{
#'   font_cache_set("my_app")
#'   shiny::runApp("my_app")
#' }
font_cache_set <- function(path, cleanup = FALSE) {
  old_home <- font_cache_housing()
  # Copy the existing cache over to the new cache
  font_files <- dir(old_home, recursive = TRUE)
  file.copy(
    file.path(old_home, font_files),
    file.path(path, font_files)
  )
  # Remove the old cache, if requested
  if (cleanup) {
    unlink(file.path(old_home, font_files), recursive = TRUE)
  }
  Sys.setenv("THEMATIC_FONT_CACHE_DIR" = path)
  invisible(old_home)
}

gfont_cache_dir <- function(family) {
  file.path(font_cache_housing(), gfont_id(family))
}

gfont_id <- function(family) {
  gsub("\\s+", "-", tolower(family))
}

font_cache_housing <- function() {
  Sys.getenv(
    "THEMATIC_FONT_CACHE_DIR",
    file.path(thematic_cache_dir(), "fonts")
  )
}


thematic_cache_dir <- function() {
  rappdirs::user_cache_dir("thematic")
}

# Apparently the only way to set lattice pars in a way that is independent
# of graphics device state is to attach those pars to the lattice object...
# This excerpt is taken from the @seealso section of help(trellis.device)
#>
#> trellis.par.get and trellis.par.set can be used to query and modify the
#> settings after a device has been initialized. The par.settings argument to
#> high level functions, described in xyplot, can be used to attach transient
#> settings to a "trellis" object.

lattice_print_set <- function(theme) {
  if (!is_installed("lattice")) return(NULL)
  lattice_print_restore()
  .globals$lattice_print <- lattice::lattice.getOption("print.function") %||%
    getFromNamespace("plot.trellis", "lattice")
  lattice::lattice.options(
    print.function = function(x, ...) {
      # Force our before.grid.newpage hook to happen *before*
      # lattice's print function is called, which needs to happen
      # in order for us to resolve the font *before* drawing
      grid::grid.newpage()
      x$par.settings <- lattice_par()
      .globals$lattice_print(x, ..., newpage = FALSE)
    }
  )
}

lattice_print_restore <- function() {
  if (!exists("lattice_print", envir = .globals)) return()
  lattice::lattice.options(print.function = .globals$lattice_print)
  rm("lattice_print", envir = .globals)
}

lattice_par <- function() {
  theme <- .globals$theme
  bg <- theme$bg
  fg <- theme$fg
  font <- theme$font

  params <- list(
    # See figure 9.3 for an example of where grid gpar matters
    # http://lmdvr.r-forge.r-project.org/figures/figures.html
    grid.pars =         list(col = fg, fontfamily = font$family, cex = font$scale),
    background =        list(col = bg),
    reference.line =    list(col = bg),
    panel.background =  list(col = mix_colors(theme$bg, theme$fg, 0.1)),
    strip.background =  list(col = mix_colors(theme$bg, theme$fg, 0.2)),
    strip.border =      list(col = fg),
    axis.line =         list(col = fg),
    axis.text =         list(col = fg),
    add.line =          list(col = fg),
    add.text =          list(col = fg),
    par.xlab.text =     list(col = fg),
    par.ylab.text =     list(col = fg),
    par.zlab.text =     list(col = fg),
    par.main.text =     list(col = fg),
    par.sub.text =      list(col = fg),
    box.3d =            list(col = fg),
    plot.polygon =      list(border = fg),
    superpose.polygon = list(border = fg),
    box.dot =           list(col = fg),
    dot.line =          list(col = mix_colors(theme$bg, theme$fg, 0.2))
  )

  # For lattice, accent can be of length 2, one to specify
  # 'stroke' accent and one for fill accent
  accent <- rep(theme$accent, length.out = 2)
  if (sum(is.na(accent)) == 0) {
    params$plot.line$col <-     accent[[1]]
    params$plot.symbol$col <-   accent[[1]]
    params$dot.symbol$col <-    accent[[1]]
    params$box.rectangle$col <- accent[[1]]
    params$box.umbrella$col <-  accent[[1]]
    params$plot.polygon$col <-  accent[[2]]
    params$grid.pars$fill <-    accent[[2]]
  }

  qualitative <- theme$qualitative
  if (sum(is.na(qualitative)) == 0) {
    # I'm not in love with the idea of this; but alas, it's consistent with lattice's default
    region_pal <- colorRampPalette(c(qualitative[[1]], bg, qualitative[[2]]))
    params$strip.shingle$col <-     qualitative
    params$regions$col <-           region_pal(100)
    params$superpose.line$col <-    qualitative
    params$superpose.symbol$col <-  qualitative
    params$superpose.symbol$fill <- qualitative
    params$superpose.polygon$col <- qualitative
  }

  params
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom graphics par plot plot.new
#' @importFrom grid grid.newpage grid.draw grid.text gpar seekViewport pushViewport upViewport
#' @importFrom grDevices col2rgb dev.new dev.cur dev.off dev.set devAskNewPage palette colorRampPalette
#' @importFrom utils capture.output download.file unzip packageVersion modifyList getFromNamespace getS3method assignInNamespace
#' @importFrom rlang is_installed list2 %|%
#' @importFrom rstudioapi getThemeInfo versionInfo readRStudioPreference
## usethis namespace: end
NULL

# Logic for adjusting a color based on bg/fg/accent
adjust_color <- function(color, bg, fg, accent = NA) {
  if (!length(color)) return(color)
  if (length(color) > 1) {
    stop("Internal error: adjust_color() expects an input of length 1")
  }
  if (is.na(color) || identical(color, "NA") || color == "0") return(color)

  # If a gray scale color, then the degree of gray determines
  # the mixing between fg (aka black) and bg (aka white)
  rgbs <- col2rgb(color, alpha = TRUE)[1:3,1]
  if (sum(diff(rgbs)) == 0) {
    return(mix_colors(bg, fg, 1 - (rgbs[1] / 255)))
  }

  # At this point we should be dealing with an accent color...
  # If accent is NA though, then the user has specified to NOT change it
  if (is.na(accent)) color else accent
}

mix_colors <- function(bg, fg, amount) {
  if (!length(bg) || !length(fg)) return(NULL)
  scales::colour_ramp(c(bg, fg), alpha = TRUE)(amount)
}

# Estimate the amount of mixture `bg` & `fg` is takes to get `color`
# Note that method = 'cmc' gives much better results than 'cie2000'
# for the important case of (theme_gray()$panel.background):
# scales::show_col(c(mix_colors("white", "black", amount_of_mixture("gray92", "white", "black")), "gray92"))
amount_of_mixture <- function(color, bg, fg) {
  bg_dist <- farver::compare_colour(
    farver::decode_colour(color), farver::decode_colour(bg),
    from_space = "rgb", method = "cmc"
  )
  fg_dist <- farver::compare_colour(
    farver::decode_colour(color), farver::decode_colour(fg),
    from_space = "rgb", method = "cmc"
  )
  as.numeric(bg_dist / (bg_dist + fg_dist))
}

# x should be of length 1
parse_any_color <- function(x) {
  if (length(x) != 1) {
    stop("Internal thematic error. parse_any_color() should be used on length 1 input")
  }
  y <- tryCatch(
    col2rgb(x),
    error = function(e) {
      y <- htmltools::parseCssColors(x, mustWork = FALSE)
      if (is.na(y)) stop("Invalid color specification '", x, "'.", call. = FALSE)
      y
    }
  )
  if (is.character(y)) y else x
}

download_file <- function(url, dest, ...) {
  if (is_installed("curl")) {
    if (!curl::has_internet()) {
      warning(
        "Looks like you don't have internet access, which is needed to ",
        "download and install Google Fonts files. Try either changing ",
        "thematic::font_spec(), manually installing the relevant font, or ",
        "trying again with internet access.",
        call. = FALSE
      )
    }
    return(curl::curl_download(url, dest, ...))
  }

  if (capabilities("libcurl")) {
    return(download.file(url, dest, method = "libcurl", ...))
  }

  stop(
    "Downloading Google Font files requires either the curl package or ",
    "`capabilities('libcurl')`. ", call. = FALSE
  )
}

is_rstudio <- function(version_needed = NULL) {
  rstudioapi::isAvailable(version_needed)
}

in_rstudio_gd <- function(dev_name = infer_device()) {
  "RStudioGD" %in% dev_name
}

try_get_theme_info <- function() {
  if (!is_rstudio()) return(NULL)
  if (!rstudioapi::hasFun("getThemeInfo")) return(NULL)

  tryCatch(
    getThemeInfo(),
    error = function(err) {
      warning(
        "Could not get current IDE theme info: ",
        conditionMessage(err),
        call. = FALSE
      )
      NULL
    }
  )
}

# If the current device is null, try to open the default device
# infer what it'll be
infer_device <- function() {
  if (!is_null_device()) {
    return(.Device)
  }
  dev <- attempt_with_new_device(.Device)
  if (!is.null(dev) && !is_null_device(dev)) {
    return(dev)
  }
  # In this case, the system's default device isn't supported,
  # but it could be that a device might be available
  # Attempt to open the default device and ask for its name
  dev <- attempt_with_device(default_device(), {.Device})
  if (!is.null(dev) && !is_null_device(dev)) {
    return(dev)
  }
  stop(
    "It seems your system doesn't support an R graphics device. ",
    "Try installing the ragg and/or Cairo packages.",
    call. = FALSE
  )
}

# .Device is equivalent to names(dev.cur())
is_null_device <- function(x = .Device) {
  identical(x, "null device")
}

attempt_with_new_device <- function(expr) {
  attempt_with_device(dev_new, expr)
}

attempt_with_device <- function(dev_fun, expr, fail_value = NULL) {
  tmp <- tempfile()
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  dev_before <- dev.cur()
  if (!is.function(dev_fun)) {
    stop("Internal error: dev_fun should be a function.")
  }
  file_arg <- grep("^file", names(formals(dev_fun)), value = TRUE)
  if (length(file_arg) == 0) {
    stop("Internal error: expect graphics device function to have a file/filename argument.")
  }
  res <- try(do.call(dev_fun, rlang::set_names(list(tmp), file_arg[1])))
  if (inherits(res, "try-error")) {
    maybe_warn(
      "thematic tried but failed to open a graphics device. If plots don't render ",
      "how you'd expect them to, try setting `options(device = ...)` to a device ",
      "that is supported on your system (e.g., `png`, `jpeg`, `Cairo::Cairo`, etc).",
      id = "no-graphics-device"
    )
    return(fail_value)
  }

  # dev.off() closes the current device, then sets the current
  # device to the _next_ device, which isn't necessarily the
  # previously open device.
  dev_after <- dev.cur()
  on.exit({
    dev.off(dev_after)
    # This next line is here to avoid this situation
    # > png(); png(); png(); dev.list()
    # quartz_off_screen quartz_off_screen quartz_off_screen
    # 2                 3                 4
    # > dev.off(); dev.cur()
    # quartz_off_screen
    # 2
    if (dev_before > 1) dev.set(dev_before)
  }, add = TRUE)

  force(expr)
}

dev_new <- function(filename) {
  # If this is called via thematic_save_plot(), then we know
  # exactly what function and args to use to clone the device
  if (length(.globals$device)) {
    do.call(.globals$device$fun, .globals$device$args)
    return()
  }
  # Most devices use `filename` instead of `file`,
  # but there are a few exceptions (e.g., pdf(), svglite::svglite())
  dev.new(filename = filename, file = filename)
}




dropNulls <- function(x) {
  x[!vapply(x, is.null, FUN.VALUE=logical(1))]
}

tryNULL <- function(expr) {
  tryCatch(expr, error = function(e) NULL)
}

tryGet <- function(...) {
  tryCatch(get(...), error = function(e) NULL)
}

"%||%" <- function(x, y) {
  if (!length(x)) y else x
}

"%OR%" <- function(x, y) {
  if (is.null(x) || isTRUE(is.na(x))) y else x
}

library(shiny)
library(thematic)

thematic_shiny(font = "auto")
# https://github.com/rstudio/thematic/pull/68
opts <- options(device = function() stop("boom"))
onStop(function() options(opts))

ui <- fluidPage(
  titlePanel("Hello"),
  tags$head(
    tags$link(href="https://fonts.googleapis.com/css?family=Pacifico", rel="stylesheet"),
    tags$style(HTML("body{background-color:#444444; color:#e4e4e4; font-family: Pacifico}")),
    tags$style(HTML("a{color:#e39777}"))
  ),
  plotOutput("p")
)

render_plot <- function(expr) {
  snapshotPreprocessOutput(renderPlot(expr), function(value) {})
}

server <- function(input, output, session) {
  output$p <- render_plot({
    ggplot2::qplot(1:10, 1:10, color = 1:10)
  })
}

shinyApp(ui, server)

library(shiny)
library(ggplot2)
library(thematic)

# The device we're testing
device <- "quartz_png"
ext <- "png"

ui <- fluidPage(
  imageOutput("ggplot"),
  imageOutput("lattice"),
  imageOutput("base")
)

render_image <- function(expr) {
  snapshotPreprocessOutput(renderImage(expr, deleteFile = FALSE), function(value) {})
}

image_info <- function(type, ext) {
  list(
    src = paste0(type, "-1.", ext),
    width = 600, height = 400
  )
}

server <- function(input, output, session) {
  # Knit all the Rmds
  # Note that this is done inside the server function
  # to avoid a timeout issue with shinytest::testApp()
  rmd_txt <- knitr::knit_expand("../template_device.Rmd", device = sprintf("'%s'", device))
  res <- callr::r(function(...) { knitr::knit2html(...) }, args = list(text = rmd_txt))

  output$ggplot <- render_image(image_info("ggplot", ext))
  output$lattice <- render_image(image_info("lattice", ext))
  output$base <- render_image(image_info("base", ext))

  onFlush(function() {
    unlink(dir(pattern = paste0("\\.", ext)))
  })
}

shinyApp(ui, server)

library(shiny)
library(ggplot2)
library(thematic)

thematic_shiny("black", "white", font = font_spec("Pacifico", 1.25, update = TRUE))

ui <- fluidPage(
  imageOutput("quartz")
)

server <- function(input, output, session) {
  output$quartz <- snapshotPreprocessOutput(renderImage({
    file <- thematic_save_plot(
      qplot(1:10) + ggtitle("quartz png"),
      device = grDevices::png
    )
    list(src = file, width = 480, height = 480)
  }), function(value) {})
}

shinyApp(ui, server)

try_gfont_download_and_register <- function(family, quiet = TRUE, systemfonts = FALSE) {
  # If no match in Google Font cache, look for a match with an updated set
  google_fonts <- get_google_fonts()
  family_idx <- match(family, google_fonts$family)

  if (is.na(family_idx)) {
    maybe_warn(
      "The font family '", family, "' doesn't appear to be available as a ",
      "Google Font. Try manually downloading and installing it on your system. ",
      "For more info, visit https://github.com/rstudio/thematic#fonts",
      id = paste0(family, "-not-a-google-font")
    )
    return(FALSE)
  }

  font_info <- google_fonts[family_idx, , drop = TRUE]

  download_files <- function(urls, dests) {
    urls <- urls[!is.na(urls)]
    dests <- dests[!is.na(urls)]
    Map(function(x, y) download_file(x, y, quiet = quiet), urls, dests)
  }

  cache_dir <- gfont_cache_dir(family)
  if (!dir.exists(cache_dir)) dir.create(cache_dir, recursive = TRUE)

  # Columns pointin to the relevant face-specific ttf files
  # (these columns are added by add_gfont_faces())
  faces <- c("regular", "bold", "italic", "bolditalic")
  dests <- file.path(cache_dir, paste0(faces, ".ttf"))

  download_files(font_info[faces], dests)
  # Store the font info with the cache so we know if it needs
  # to be updated later on
  saveRDS(font_info, file = file.path(cache_dir, "cache.rds"))
  try_register_gfont_cache(family, systemfonts)
}


try_register_gfont_cache <- function(family, systemfonts = FALSE) {
  google_fonts <- get_google_fonts()
  family_idx <- match(family, google_fonts$family)
  if (is.na(family_idx)) return()

  cache_dir <- gfont_cache_dir(family)
  if (!dir.exists(cache_dir)) return()

  font_file <- function(variant) {
    dir(
      cache_dir, full.names = TRUE,
      pattern = paste0("^", variant, "\\.ttf$")
    )
  }

  regular <- font_file("regular")
  bold <- font_file("bold")
  italic <- font_file("italic")
  bolditalic <- font_file("bolditalic")

  if (systemfonts) {
    tryCatch(
      systemfonts::register_font(
        family, regular,
        bold = bold,
        italic = italic,
        bolditalic = bolditalic
      ),
      error = function(e) {
        msg <- conditionMessage(e)
        # It's intentional that we don't check `family %in% system_font()` before
        # registering since that's a non-trivial cost that must also be paid in
        # `register_font()`. So, if the failure is due to a system font already
        # being available, then don't throw since that's irrelevant to the end user
        # https://github.com/r-lib/systemfonts/blob/e4a98b/R/register_font.R#L60
        if (!grepl("system font with that family name", msg)) {
          warning(msg)
        }
      }
    )
  } else if (is_installed("sysfonts")) {
    tryCatch(
      sysfonts::font_add(
        family, regular,
        bold = bold,
        italic = italic,
        bolditalic = bolditalic
      ),
      error = function(e) {
        warning(conditionMessage(e))
      }
    )
  }

}

get_google_fonts <- function() {
  .globals$google_fonts %||% google_fonts
}

update_gfonts <- function() {
  pkgs <- c("curl", "jsonlite")
  for (pkg in pkgs) {
    if (is_installed(pkg)) next
    stop("Using an updated set of google fonts requires the ", pkg, "package to be installed", call. = FALSE)
  }
  .globals$google_fonts <- tryCatch(
    add_gfont_faces(jsonlite::fromJSON(gfont_api_url())$items),
    error = function(e) {
      maybe_warn(
        "Failed to update google fonts",
        id = "update-gfonts-failed"
      )
      NULL
    }
  )
}

# Do our best to map font-weight to R's font faces
add_gfont_faces <- function(google_fonts) {
  files <- google_fonts$files
  google_fonts$regular <- files$regular %|% files$`300`

  google_fonts$bold <- files$`700` %|%
    files$`800` %|% files$`600` %|%
    files$`900` %|% files$`500`

  google_fonts$italic <- files$italic %|% files$`300italic`

  google_fonts$bolditalic <- files$`700italic` %|%
    files$`800italic` %|% files$`600italic` %|%
    files$`900italic` %|% files$`500italic`

  google_fonts
}

update_gfonts_cache <- function(quiet = FALSE) {
  font_dirs <- list.dirs(font_cache_housing(), full.names = TRUE, recursive = FALSE)
  gfonts <- get_google_fonts()
  for (font in font_dirs) {
    cache <- readRDS(file.path(font, "cache.rds"))
    lastModified <- gfonts[match(cache$family, gfonts$family), "lastModified", drop = TRUE]
    if (isTRUE(lastModified > cache$lastModified)) {
      if (!quiet) message("Updating out-of-date font cache for family '", cache$family, "'")
      try_gfont_download_and_register(cache$family, quiet)
    }
  }
}

gfont_api_url <- function() {
  paste0(
    "https://www.googleapis.com/webfonts/v1/webfonts?key=",
    gfont_key()
  )
}

gfont_key <- function() {
  Sys.getenv(
    "GFONT_KEY",
    paste0("AIzaSyDP", "KvElVqQ-", "26f7tjxyg", "IGpIajf", "tS_zmas")
  )
}

library(shiny)
library(ggplot2)
library(thematic)

# The device we're testing
ext <- "png"

ui <- fluidPage(
  imageOutput("ggplot"),
  imageOutput("lattice"),
  imageOutput("base")
)

render_image <- function(expr) {
  snapshotPreprocessOutput(renderImage(expr, deleteFile = FALSE), function(value) {})
}

image_info <- function(type, ext) {
  list(
    src = paste0(type, "-1.", ext),
    width = 600, height = 400
  )
}

# Knit all the Rmds
# Note that this is done inside the server function
# to avoid a timeout issue with shinytest::testApp()
infile <- "darkly.Rmd"
rmd_txt <- knitr::knit_expand("../../template_theme.Rmd", theme = "darkly")
writeLines(rmd_txt, infile)
outfile <- rmarkdown::render(infile)

onStop(function() {
  unlink(infile)
  unlink(dir(pattern = paste0("\\.", ext)))
  unlink(
    c(infile, outfile, paste0(tools::file_path_sans_ext(infile), "_files")),
    recursive = TRUE
  )
})

server <- function(input, output, session) {

  output$ggplot <- render_image(image_info("ggplot", ext))
  output$lattice <- render_image(image_info("lattice", ext))
  output$base <- render_image(image_info("base", ext))
}

shinyApp(ui, server)

# TODO: .globals is currently used all over the place. Provide an official getter/setter
.globals <- new.env(parent = emptyenv())

.warning_env <- new.env(parent = emptyenv())

# Inspired by lifecycle:::needs_warning()
# https://github.com/r-lib/lifecycle/blob/ef16c98fb4/R/signal-deprecated.R#L173
needs_warning <- function(id) {
  last <- .warning_env[[id]]
  if (is.null(last)) {
    return(TRUE)
  }

  if (!inherits(last, "POSIXct")) {
    stop("Internal error: Expected `POSIXct` value in `needs_warning()`.", call. = FALSE)
  }

  # Warn every 8 hours
  (Sys.time() - last) > (8 * 60 * 60)
}


maybe_warn <- function(..., id, call. = FALSE) {
  if (needs_warning(id)) {
    .warning_env[[id]] <- Sys.time()
    warning(..., call. = call.)
  }
}

install_knitr_hooks <- function() {
  if (!is_installed("knitr")) {
    return()
  }

  # Try our best to set better chunk defaults for the fig.showtext and dev.args options...
  # Unfortunately, both of these options must be set prior to code evaluation to take effect,
  # but we won't know the "true" value that these options should be until after evaluation
  knitr::opts_chunk$set(thematic_hook = TRUE)
  knitr::opts_hooks$set(thematic_hook = function(options) {
    # Default to using showtext, but allow users to opt-out if they don't need it
    options$fig.showtext <- getOption("thematic.fig.showtext", is_installed("showtext"))
    bg <- thematic_get_option("bg", "white", resolve = TRUE)
    dev <- options[["dev"]]
    # dev.args can also be a list of lists of args
    if (is.character(dev) && isTRUE(dev %in% names(options$dev.args))) {
      options$dev.args[[dev]]$bg <- bg
    } else {
      options$dev.args$bg <- bg
    }
    options
  })
}

library(shiny)
library(ggplot2)
library(thematic)

thematic_shiny("black", "white", font = font_spec("Pacifico", 1.25, update = TRUE))

ui <- fluidPage(
  imageOutput("png")
)

render_image <- function(expr) {
  snapshotPreprocessOutput(renderImage(expr), function(value) {})
}

server <- function(input, output, session) {
  output$png <- render_image({
    file <- thematic_save_plot(
      qplot(1:10) + ggtitle("grDevices::png"),
      device = grDevices::png, type = "cairo",
      antialias = "none",
      width = 800, height = 400
    )
    list(src = file, width = 800, height = 400)
  })
}

shinyApp(ui, server)

# -----------------------------------------------------------------------------------
# Custom ggplot_build()
#
# N.B. Overriding the build method is obviously not ideal, but compared to alternative
# approaches to setting/restoring Geom and Scale defaults, it seems like the best
# option at the moment. A future version of ggplot2 should provide sufficient tools for
# managing scale defaults management, but it's unclear whether we'll have the same
# for Geoms (this PR might do it, but it might not https://github.com/tidyverse/ggplot2/pull/2749)
# -----------------------------------------------------------------------------------

ggplot_build_set <- function() {
  if (!is_installed("ggplot2")) return(NULL)
  # Alternative update method if we're dealing with S7
  if ("class_ggplot" %in% getNamespaceExports("ggplot2")) {
    .globals$ggplot_build <- .globals$ggplot_build %||%
      getS3method("ggplot_build", "ggplot2::ggplot", envir = asNamespace("ggplot2"))
    registerS3method("ggplot_build", "ggplot2::ggplot", ggthematic_build, asNamespace("ggplot2"))
    return(NULL)
  }
  ggplot_build_restore()
  # Note that assignInNamespace() does S3 method registration, but to
  # find the relevant generic, it looks in the parent.frame()...
  # so this line here is prevent that from failing if ggplot2 hasn't been attached
  # https://github.com/wch/r-source/blob/d0ede8/src/library/utils/R/objects.R#L472
  ggplot_build <- getFromNamespace("ggplot_build", "ggplot2")
  .globals$ggplot_build <- getFromNamespace("ggplot_build.ggplot", "ggplot2")
  assign_in_namespace <- assignInNamespace
  ensure_s3_methods_matrix()
  assign_in_namespace("ggplot_build.ggplot", ggthematic_build, "ggplot2")
  if (is_installed("gganimate")) {
    .globals$gganim_build <- getFromNamespace("ggplot_build.gganim", "gganimate")
    formals(ggthematic_build)$ggplot_build <- .globals$gganim_build
    assign_in_namespace("ggplot_build.gganim", ggthematic_build, "gganimate")
  }
}

ggplot_build_restore <- function() {
  # Alternative update method if we're dealing with S7
  if (is.function(.globals$ggplot_build)) {
    if ("class_ggplot" %in% getNamespaceExports("ggplot2")) {
      registerS3method("ggplot_build", "ggplot2::ggplot", .globals$ggplot_build, asNamespace("ggplot2"))
      rm("ggplot_build", envir = .globals)
      return(NULL)
    }
    ggplot_build <- getFromNamespace("ggplot_build", "ggplot2")
    assign_in_namespace <- assignInNamespace
    ensure_s3_methods_matrix()
    assign_in_namespace("ggplot_build.ggplot", .globals$ggplot_build, "ggplot2")
    rm("ggplot_build", envir = .globals)
    if (is.function(.globals$gganim_build)) {
      assign_in_namespace("ggplot_build.gganim", .globals$gganim_build, "gganimate")
      rm("gganim_build", envir = .globals)
    }
  }
}

# When registerS3method() is called in an onLoad() hook, it causes
# assignInNamespace() to no longer work for S3 methods because this
# S3methods namespace info gets coerced into a list
# (which assignInNamespace() isn't expecting). Thus, here, we make
# sure that the info is a matrix before calling assignInNamespace()
# https://github.com/rstudio/thematic/issues/90#issuecomment-780224962
ensure_s3_methods_matrix <- function(pkg = "ggplot2") {
  S3 <- .getNamespaceInfo(asNamespace(pkg), "S3methods")
  S3 <- matrix(as.character(S3), nrow = nrow(S3), ncol = ncol(S3))
  setNamespaceInfo(asNamespace(pkg), "S3methods", S3)
}

# N.B. If you make changes here, plotly might have to as well!
# https://github.com/ropensci/plotly/pull/1801/files#diff-3afd3a8e6a2cbc84a7afc6d2d06ec5e3R429
ggthematic_build <- function(p, ggplot_build = NULL, theme = NULL) {
  theme <- theme %||% thematic_get_theme(resolve = TRUE)
  ggplot_build <- ggplot_build %||% .globals$ggplot_build
  if (!is.function(ggplot_build)) {
    stop("`ggplot_build` must be a function", call. = FALSE)
  }
  if (!length(theme)) {
    return(ggplot_build(p))
  }

  fg <- theme$fg
  bg <- theme$bg
  # Accent can be of length 2 because lattice
  accent <- theme$accent[1]
  qualitative <- theme$qualitative
  sequential <- theme$sequential

  # Collect all the plot's geoms, as well as some 'core' geoms,
  # since some geoms, e.g. GeomSf, want their default_aes to derive
  # from 'lower-level' geoms, like GeomPoint, GeomLine, GeomPolygon
  geoms <- c(
    lapply(p$layers, function(x) x$geom),
    lapply(
      c("GeomPoint", "GeomLine", "GeomPolygon"),
      getFromNamespace, "ggplot2"
    )
  )

  # Remember defaults
  if ("get_geom_defaults" %in% getNamespaceExports("ggplot2")) {
    get_geom_defaults <- get("get_geom_defaults", asNamespace("ggplot2"))
  } else {
    get_geom_defaults <- function(geom, ...) geom$default_aes
  }
  user_defaults <- lapply(geoms, get_geom_defaults, theme = theme)

  # Modify defaults
  Map(function(geom, user_default) {
    ggplot2::update_geom_defaults(
      geom, list(
        colour = adjust_color(user_default$colour, bg, fg, accent),
        fill = adjust_color(user_default$fill, bg, fg, accent)
      )
    )

    # i.e., GeomText/GeomLabel
    if ("family" %in% names(geom$default_aes)) {
      ggplot2::update_geom_defaults(
        geom, list(
          family = theme$font$family,
          size = user_default$size * theme$font$scale
        )
      )
    }
  }, geoms, user_defaults)

  # Restore users defaults on exit
  on.exit({
    Map(function(x, y) { x$default_aes <- y }, geoms, user_defaults)
  }, add = TRUE)

  # Modify scaling defaults
  scale_defaults <- list()
  if (!is_na_scalar(qualitative)) {
    scale_defaults$ggplot2.discrete.colour <- qualitative
    scale_defaults$ggplot2.discrete.fill <- qualitative
  }
  if (!is_na_scalar(sequential)) {
    scale_defaults$ggplot2.continuous.colour <- function(...) {
      ggplot2::scale_colour_gradientn(..., colours = sequential)
    }
    scale_defaults$ggplot2.continuous.fill <- function(...) {
      ggplot2::scale_fill_gradientn(..., colours = sequential)
    }
    scale_defaults$ggplot2.binned.colour <- function(...) {
      ggplot2::scale_colour_stepsn(..., colours = sequential)
    }
    scale_defaults$ggplot2.binned.fill <- function(...) {
      ggplot2::scale_fill_stepsn(..., colours = sequential)
    }
  }

  # If we have modern ggplot2, use the official scale default APIs.
  if (has_proper_ggplot_scale_defaults()) {
    old_scales <- do.call(options, scale_defaults)
    on.exit({options(old_scales)}, add = TRUE)
  } else {
    # This isn't an officially supported way of setting default scales, but
    # `scales_add_defaults()` first looks in the plot_env to find default scales
    # https://github.com/tidyverse/ggplot2/blob/a7b3135/R/layer.r#L214
    if (!is_na_scalar(sequential)) {
      if (!p$scales$has_scale("colour")) {
        # TODO: distinguish between existing/NULL
        colour_continuous <- tryGet("scale_colour_continuous", envir = p$plot_env)
        assign("scale_colour_continuous", scale_defaults$ggplot2.continuous.colour, envir = p$plot_env)
        on.exit({
          restore_scale("scale_colour_continuous", colour_continuous, envir = p$plot_env)
        }, add = TRUE)
      }
      if (!p$scales$has_scale("fill")) {
        fill_continuous <- tryGet("scale_fill_continuous", envir = p$plot_env)
        assign("scale_fill_continuous", scale_defaults$ggplot2.continuous.fill, envir = p$plot_env)
        on.exit({
          restore_scale("scale_fill_continuous", fill_continuous, envir = p$plot_env)
        }, add = TRUE)
      }
    }
    if (!is_na_scalar(qualitative)) {
      if (!p$scales$has_scale("colour")) {
        colour_discrete <- tryGet("scale_colour_discrete", envir = p$plot_env)
        assign(
          "scale_colour_discrete",
          function(...) ggplot2::discrete_scale("colour", "qualitative", qualitative_pal(qualitative), ...),
          envir = p$plot_env
        )
        on.exit({
          restore_scale("scale_colour_discrete", colour_discrete, envir = p$plot_env)
        }, add = TRUE)
      }

      if (!p$scales$has_scale("fill")) {
        fill_discrete <- tryGet("scale_fill_discrete", envir = p$plot_env)
        assign(
          "scale_fill_discrete",
          function(...) ggplot2::discrete_scale("fill", "qualitative", qualitative_pal(qualitative), ...),
          envir = p$plot_env
        )
        on.exit({
          restore_scale("scale_fill_discrete", fill_discrete, envir = p$plot_env)
        }, add = TRUE)
      }
    }
  }

  # Since thematic_theme() wants to define elements that could possibly be
  # parents of the user's theme elements, we iterate through each generation of the
  # tree and merge the user theme elements with thematic_theme()
  theme_final <- theme_thematic(theme)
  theme_user <- resolve_theme_inheritance(p$theme)
  for (name in names(theme_user)) {
    theme_final[[name]] <- ggplot2::merge_element(
      new = theme_user[[name]], old = theme_final[[name]]
    )
  }

  # There are some theme settings that are best not filled in yet.
  # These gain different internal defaults based on the text position and
  # setting them here would disrupt that mechanism.
  restore <- c("hjust", "vjust", "margin")
  for (name in c("title", "legend.title", "legend.text")) {
    if (ggplot2::is_theme_element(theme_final[[name]], "blank")) {
      next
    }
    if (ggplot2::is_theme_element(p$theme[[name]], "blank")) {
      value <- list(NULL)
    } else {
      value <-  p$theme[[name]][restore] %||% list(NULL)
    }
    theme_final[[name]][restore] <- value
  }
  p$theme <- theme_final

  ggplot_build(p)
}

# ----------------------------------------------------------------------
# Returns a modified version of the global theme based on thematic theme
# N.B. if and when ggplot2 gets the ability to set geom and scale defaults
# from a theme object, then it might make sense to export this function
# ----------------------------------------------------------------------
theme_thematic <- function(theme = .globals$theme) {
  # Calculate elements so we know the "final form" of each element
  # (i.e., takes care of inheritance)
  ggtheme <- computed_theme_elements(ggplot2::theme_get())

  # Handles any missing color value (e.g., NULL, NA, 'transparent')
  `%missing%` <- function(x, y) {
    if (identical(x, "transparent")) return(y)
    x %OR% y
  }

  # The remaining updates depend on the old fg/bg
  old_bg <- ggtheme$plot.background$fill %missing% .globals$base_params$bg %missing% "white"
  old_fg <- ggtheme$title$colour %missing% "black"

  new_fg <- theme$fg
  new_bg <- theme$bg

  # Main idea: theme colors are comprised of mixtures of bg and fg
  # So, given a color, get it's perceptual distance between bg <-> fg
  # and use that 'amount of mixture' to inform a new color
  update_color <- function(color) {
    amt <- amount_of_mixture(color, old_bg, old_fg)
    mix_colors(new_bg, new_fg, amt)
  }

  update_element <- function(element, name) {
    UseMethod("update_element")
  }

  update_element.element_text <- function(element, name) {
    ggplot2::element_text(
      colour = update_color(element$colour),
      size = element$size * theme$font$scale,
      family = if (!identical(theme$font$family, "")) theme$font$family
    )
  }

  update_element.element_rect <- function(element, name) {
    ggplot2::element_rect(
      fill = update_color(element$fill),
      colour = update_color(element$colour)
    )
  }

  update_element.element_line <- function(element, name) {
    ggplot2::element_line(
      colour = update_color(element$colour)
    )
  }

  update_element.element_blank <- function(element, name) {
    # Make sure plot.background is always defined; since otherwise,
    # we'd have to depend on par("bg") being set (and the device respecting it)
    if (name %in% c("plot.background", "panel.background")) {
      ggplot2::element_rect(fill = new_bg, colour = new_bg)
    } else {
      element
    }
  }

  update_element.default <- function(element, name) NULL

  ggtheme <- Map(function(x, y) update_element(x, y), ggtheme, names(ggtheme))
  do.call(ggplot2::theme, dropNulls(ggtheme))
}

# Get all the computed theme elements from a given theme definition
computed_theme_elements <- function(ggtheme) {
  theme_default <- ggplot2::theme_grey()
  elements <- names(ggplot2::get_element_tree())
  # If this isn't a complete theme, make it one
  # (fixes cases like ggthemes::theme_pander() which isn't complete)
  if (identical(attr(ggtheme, "complete"), FALSE)) {
    ggtheme <- theme_default + ggtheme
  }
  computed <- rlang::set_names(lapply(elements, calc_element_safe, ggtheme), elements)
  dropNulls(computed)
}


resolve_theme_inheritance <- function(p_theme) {
  if (!length(p_theme)) {
    return(p_theme)
  }
  relationships <- theme_relationships()
  while(nrow(relationships) > 0) {
    # Start from the top of the tree (i.e., elements that don't have a parent)
    # and work our way down, merging the child with the parent (if both are defined)
    idx <- !relationships$parent %in% relationships$child
    for (i in which(idx)) {
      this_relationship <- relationships[i, ]
      this_kid <- this_relationship$child
      this_parent <- this_relationship$parent
      parent_el <- p_theme[[this_parent]]
      kid_el <- p_theme[[this_kid]]
      # parent doesn't exist so do nothing
      if (is.null(parent_el)) {
        next
      }
      p_theme[[this_kid]] <- if (!is.null(kid_el)) {
        # both parent & child exist
        ggplot2::merge_element(new = kid_el, old = parent_el)
      } else {
        # parent exists but child doesnt
        parent_el
      }
    }
    # remove relationships that have been inspected
    relationships <- relationships[!idx, ]
  }
  p_theme
}

# TODO: could consider memoising if this proves to be a bottleneck
theme_relationships <- function() {
  inherits <- lapply(ggplot2::get_element_tree(), function(x) x$inherit)
  data.frame(
    child  = rep(names(inherits), lengths(inherits)),
    parent = unlist(inherits),
    stringsAsFactors = FALSE
  )
}

# It's not safe to calc all elements (e.g., plot.margin)
calc_element_safe <- function(x, ...) {
  tryCatch(ggplot2::calc_element(x, ...), error = function(e) x)
}

restore_scale <- function(name, x, envir) {
  if (is.null(x)) rm(name, envir = envir) else assign(name, x, envir = envir)
}

# Introduced scale defaults via options()
has_proper_ggplot_scale_defaults <- function() {
  packageVersion("ggplot2") >= "3.3.2"
}


qualitative_pal <- function(codes) {
  function(n) {
    if (n <= length(codes)) {
      codes[seq_len(n)]
    } else {
      scales::hue_pal()(n)
    }
  }
}

is_na_scalar <- function(x) {
  if (length(x) != 1) {
    return(FALSE)
  }
  is.na(x)
}

#' Capture a thematic plot as a saved file
#'
#' Uses a `device` to capture the result of an expression (`expr`)
#' that produces a plot. If `default_device()` is used, custom fonts
#' (specified through [font_spec()]) are guaranteed to work, as long as
#' one of either the showtext or ragg package(s) are installed.
#'
#' @return `thematic_save_plot()` returns the `filename` of the produced plot
#'   and `default_device()` returns a graphics device function.
#'
#' @param expr an expression that produces a plot.
#' @param device a graphics device to use for capturing the plot.
#' @param filename a filename for the produced plot. The file extension should
#' match the relevant `device`.
#' @param ... arguments passed along to the graphics `device`.
#' @export
#' @examples
#'
#' library(thematic)
#' font <- font_spec("Rock Salt", scale = 1.25)
#' thematic_on("black", "white", font = font)
#' file <- thematic_save_plot(plot(1:10), res = 144)
#' if (interactive()) browseURL(file)
thematic_save_plot <- function(expr, device = default_device(),
                               filename = tempfile(fileext = ".png"), ...) {
  # N.B. implementation is quite similar to htmltools::capturePlot
  if (!is.function(device)) {
    stop(call. = FALSE, "The `device` argument should be a function, e.g. `ragg::agg_png`")
  }

  isTempFile <- missing(filename)

  # collect user and device arguments
  args <- rlang::list2(filename = filename, ...)
  device_args <- names(formals(device))

  # do our best to find the background color arg
  bg_arg <- grep("^background$|^bg$", device_args, value = TRUE)[1]
  if (isTRUE(is.na(bg_arg))) {
    stop(
      "Wasn't able to detect the background color argument for the given device, ",
      "so thematic won't automatically set it for you, but you can also set it yourself ",
      "by doing `thematic_save_plot(expr, bg_color_arg = thematic_get_option('bg'))`",
      call. = FALSE
    )
  }

  showtextOpts <- if (is.numeric(args$res) && is_installed("showtext")) {
    showtext::showtext_opts(dpi = args$res)
  }

  args[[bg_arg]] <- args[[bg_arg]] %||%
    thematic_get_option("bg", "white", resolve = TRUE)

  # Handle the case where device wants `file` instead of `filename`
  # (e.g., svglite::svglite)
  if (!"filename" %in% device_args && "file" %in% device_args) {
    args$file <- args[["file"]] %||% args$filename
    args$filename <- NULL
  }

  # dev.off() closes the current device, then sets the current
  # device to the _next_ device, which isn't necessarily the
  # previously opened device. So, remember the current device now,
  # then open a new one, then explicitly set the device to the
  # previous device (so as to not cause side-effects).
  dev_cur <- dev.cur()
  do.call(device, args)
  on.exit({
    dev.off()
    if (dev_cur > 1) dev.set(dev_cur)
  }, add = TRUE)

  # Let the world know about the device's arguments, so that
  # resolve_font_family() can use them when cloning the device
  .globals$device <- list(fun = device, args = args)
  on.exit(rm("device", envir = .globals), add = TRUE)

  # Evaluate the expression
  expr <- rlang::enquo(expr)
  tryCatch(
    {
      result <- withVisible(rlang::eval_tidy(expr))
      if (result$visible) {
        capture.output(print(result$value))
      }
      filename
    },
    error = function(e) {
      try({
        # i.e., if we _know_ this is a tempfile remove it before throwing
        if (isTempFile && file.exists(filename))
          unlink(filename)
      })
      stop(e)
    },
    finally = {
      if (length(showtextOpts)) showtext::showtext_opts(showtextOpts)
    }
  )
}

#' @rdname thematic_save_plot
#' @param type the type of output format
#' @export
default_device <- function(type = c("png", "svg", "pdf", "tiff", "jpeg")) {
  type <- match.arg(type)

  if (type %in% c("png", "tiff", "jpeg") && is_installed("ragg")) {
    dev <- switch(
      type,
      png = ragg::agg_png,
      tiff = ragg::agg_tiff,
      jpeg = ragg::agg_jpeg
    )
    return(dev)
  }

  use_cairo <- is_installed("Cairo") && !capabilities("aqua")

  switch(
    type,
    png = if (use_cairo) Cairo::CairoPNG else grDevices::png,
    svg = if (use_cairo) Cairo::CairoSVG else grDevices::svg,
    pdf = if (use_cairo) Cairo::CairoPDF else grDevices::pdf,
    tiff = if (use_cairo) Cairo::CairoTIFF else grDevices::tiff,
    jpeg = if (use_cairo) Cairo::CairoJPEG else grDevices::jpeg
  )
}

#' Enable (or disable) simplified theming of R graphics.
#'
#' A unified interface for theming **ggplot2**, **base**, and **lattice** graphics
#' based on a handful of styling options. In some cases (most notably in a **shiny** runtime),
#' these options can automatically resolve to relevant CSS styles (see the "Auto theming"
#' section below).
#'
#' @section Auto theming:
#'
#' The `bg`, `fg`, `accent`, and `font` arguments all support a value of `'auto'`,
#' which are all resolved, at plot time, based on the execution environment. In a
#' **shiny** runtime, resolution of auto values should always work as expect; but
#' in other contexts, auto values may lead to wrong or surprising results. In that
#' case, auto resolution logic can be customized (see [auto_config_set()] for more details).
#'
#' @section Global vs. local theming:
#'
#' `thematic_on()` enables thematic in a global fashion (that is, it impacts all
#' future plots, up until `thematic_off()` is called). To use thematic in local fashion,
#' first create a theme with [thematic_theme()], then provide it to [thematic_with_theme()]
#' (or similar). To use thematic in a global fashion up until a **shiny**
#' app exits, use `thematic_shiny()` (which cleans up after itself once the next shiny
#' app that exits using [shiny::onStop()]). To use thematic in a global fashion up until
#' a **rmarkdown** document finishes rendering, use `thematic_rmd()`.
#'
#' @section Color values:
#'
#' Colors (e.g., `bg`, `fg`, `accent`) may be any value understood by [col2rgb()]
#' or `htmltools::parseCssColors()` (i.e., may be any valid R or CSS color string).
#'
#' @param bg a background color.
#' @param fg a foreground color.
#' @param accent a color for making certain graphical markers 'stand out'
#' (e.g., the fitted line color for [ggplot2::geom_smooth()]).
#' Can be 2 colors for lattice (stroke vs fill accent).
#' @param font a [`font_spec()`] object. If missing, font defaults are not altered.
#' @param sequential a color palette for graphical markers that encode
#' numeric values. Can be a vector of color codes or a
#' [sequential_gradient()] object.
#' @param qualitative a color palette for graphical markers that encode
#' qualitative values (won't be used in ggplot2 when the number of data
#' levels exceeds the max allowed colors). Defaults to [okabe_ito()].
#' @param inherit should non-specified values inherit from the previous theme?
#'
#' @return [thematic_theme()] returns a theme object as a list (which can be
#' activated with [thematic_with_theme()] or [thematic_set_theme()]).
#'
#' [thematic_on()], [thematic_off()], and [thematic_shiny()] all return
#' the previous global theme.
#'
#' @seealso [sequential_gradient()], [thematic_with_theme()], [thematic_save_plot()]
#' @export
#' @examples
#' # simple dark mode
#' thematic_on("black", "white")
#' plot(1:10)
#' plot(1:10, col = 1:10)
#' lattice::show.settings()
#'
#' # use any hex color string
#' thematic_on("#444444", "#e4e4e4")
#' plot(1:10)
#' plot(1:10, col = 1:10)
#' lattice::show.settings()
#'
#' # disables thematic (also restores global state)
#' thematic_off()
#' plot(1:10)
#' lattice::show.settings()
#'
#' thematic_on("darkblue", "skyblue", "orange")
#' image(volcano)
#' image(volcano, col = thematic_get_option("sequential"))
#' lattice::show.settings()
#' thematic_off()
#'
thematic_on <- function(bg = "auto", fg = "auto", accent = "auto",
                        font = NA, sequential = sequential_gradient(),
                        qualitative = okabe_ito(), inherit = FALSE) {
  old_theme <- thematic_get_theme(resolve = FALSE)
  .globals$theme <- thematic_theme(
    bg = bg, fg = fg, accent = accent,
    font = font, sequential = sequential,
    qualitative = qualitative, inherit = inherit
  )
  # Remember par("bg") now since bg can be modified by the opening of a
  # graphics device, and if that happens before plot time, it'd be too late
  # to base_restore_params()/base_set_params() at plot time
  .globals$base_params$bg <- attempt_par("bg")

  # Remove any existing hooks before registering them
  # (otherwise, repeated calls to set_hooks will keep adding them)
  remove_hooks()
  set_hooks()

  # Override ggplot build method mainly because we currently need access to
  # the plot object in order to set Geom/Scale defaults
  ggplot_build_set()
  lattice_print_set()

  invisible(old_theme)
}

#' @rdname thematic_on
#' @export
thematic_off <- function() {
  remove_hooks()

  # Removing the plot.new hooks is not enough to restore global state
  base_params_restore()
  base_palette_restore()
  ggplot_build_restore()
  lattice_print_restore()

  theme <- .globals$theme
  if (!is.null(theme)) rm("theme", envir = .globals)

  invisible(theme)
}

#' @rdname thematic_on
#' @export
thematic_theme <- function(bg = "auto", fg = "auto", accent = "auto",
                           font = NA, sequential = sequential_gradient(),
                           qualitative = okabe_ito(), inherit = FALSE) {

  # This function is called at plot time (with bg/fg/accent)
  if (is.function(sequential)) {
    sequential <- structure("", sequential_func = sequential)
  }

  new <- structure(
    list(
      bg = tag_auto(bg),
      fg = tag_auto(fg),
      accent = tag_auto(accent),
      qualitative = qualitative,
      sequential = sequential,
      font = as_font_spec(font)
    ),
    class = "thematic_theme"
  )

  # Newly _specified_ theme elements override old elements
  old <- thematic_get_theme(resolve = FALSE)
  if (isTRUE(inherit) && length(old)) {
    fmls <- formals(thematic_theme)
    if (identical(bg, fmls$bg)) new$bg <- NULL
    if (identical(fg, fmls$fg)) new$fg <- NULL
    if (identical(accent, fmls$accent)) new$accent <- NULL
    if (identical(sequential, fmls$sequential)) new$sequential <- NULL
    if (identical(qualitative, fmls$qualitative)) new$qualitative <- NULL
    new <- utils::modifyList(old, new)
    new <- structure(new, class = "thematic_theme")
  }

  new
}

is_thematic_theme <- function(x) {
  inherits(x, "thematic_theme")
}

#' @rdname thematic_on
#' @inheritParams thematic_on
#' @param session see [shiny::onStop()].
#' @export
thematic_shiny <- function(bg = "auto", fg = "auto", accent = "auto",
                           font = NA, sequential = sequential_gradient(),
                           qualitative = okabe_ito(), inherit = FALSE,
                           session = shiny::getDefaultReactiveDomain()) {
  old_theme <- thematic_on(
    bg = bg, fg = fg, accent = accent,
    font = font, sequential = sequential,
    qualitative = qualitative, inherit = inherit
  )
  shiny::onStop(function() thematic_set_theme(old_theme), session = session)
  invisible(old_theme)
}


#' @rdname thematic_on
#' @inheritParams thematic_on
#' @export
thematic_rmd <- function(bg = "auto", fg = "auto", accent = "auto",
                         font = NA, sequential = sequential_gradient(),
                         qualitative = okabe_ito(), inherit = FALSE) {
  old_theme <- thematic_on(
    bg = bg, fg = fg, accent = accent,
    font = font, sequential = sequential,
    qualitative = qualitative, inherit = inherit
  )
  document_hook <- knitr::knit_hooks$get("document")
  knitr::knit_hooks$set(document = function(x) {
    if (is_shiny_runtime()) {
      shiny::onStop(function() thematic_set_theme(old_theme))
    } else {
      thematic_set_theme(old_theme)
    }
    document_hook(x)
  })
  invisible(old_theme)
}

is_shiny_runtime <- function() {
  isTRUE(grepl("^shiny", knitr::opts_knit$get("rmarkdown.runtime")))
}

#' Tools for getting and restoring global state
#'
#' These functions are helpful for getting and/or temporarily activating a
#' [thematic_theme()].
#'
#' @describeIn thematic_with_theme similar to [thematic_on()], but for an single plot.
#'
#' @return the result of `expr`.
#'
#' @param theme a [thematic_theme()] object.
#' @param expr R code that produces a plot.
#' @param default a default value to return in the event no thematic theme is active.
#' @export
#' @examples
#'
#' # Use thematic_with_theme() for a one-time use of thematic
#' thematic_with_theme(
#'   thematic_theme("darkblue", "skyblue", accent = "red"),
#'   plot(1:10, col = thematic_get_option("accent"), pch = 19)
#' )
#'
#' # Use thematic_set_theme() if doing something more complicated
#' # like programming on top thematic (without causing side effects)
#' my_plot <- function(expr, las = 3, ...) {
#'   old_theme <- thematic_on("black", "white")
#'   on.exit(thematic_set_theme(old_theme), add = TRUE)
#'   opts <- par(las = las)
#'   on.exit(par(opts), add = TRUE)
#'   # Imagine some more customization with ...
#'   force(expr)
#' }
#' my_plot(plot(1:10))
#'
#' thematic_off()
#' thematic_get_option("bg", "white")
#' thematic_on(bg = "red")
#' thematic_get_option("bg", "white")
#' thematic_off()
#'
#' thematic_with_theme(
#'   thematic_theme("darkblue", "skyblue"),
#'   scales::show_col(thematic_get_mixture(seq(0, 1, by = 0.1)))
#' )
#'
thematic_with_theme <- function(theme, expr) {
  old_theme <- thematic_set_theme(theme)
  on.exit(thematic_set_theme(old_theme), add = TRUE)
  # This is here so that users don't have to explictly call print()
  # on ggplot2/lattice objects (e.g., `renderPlot(thematic_with_theme(theme, qplot(1:10)))`)
  expr <- rlang::enquo(expr)
  result <- withVisible(rlang::eval_tidy(expr))
  # Don't print htmlwidgets if this happens to be used with, for example, ggplotly()
  if (result$visible && !inherits(result$value, "htmlwidget")) {
    print(result$value)
  }
  invisible(result$value)
}

#' @describeIn thematic_with_theme similar to [thematic_with_theme()], but de-couples
#'   the theme from the plot expression.
#' @param .local_envir The environment to use for scoping.
#' @export
thematic_local_theme <- function(theme, .local_envir = parent.frame()) {
  old_theme <- thematic_set_theme(theme)
  withr::defer(thematic_set_theme(old_theme), envir = .local_envir)
  invisible(old_theme)
}

#' @describeIn thematic_with_theme set a given `theme` object as the current theme.
#' @param theme a `thematic_theme()` object (or a return value of [thematic_on]/[thematic_get_theme()])
#' or `NULL` (in which case `thematic_off()` is called).
#' @export
thematic_set_theme <- function(theme) {
  if (!length(theme)) {
    return(thematic_off())
  }
  if (!is_thematic_theme(theme)) {
    stop("`theme` must be a value returned by thematic_on() or thematic_get_theme().", call. = FALSE)
  }
  do.call(thematic_on, theme)
}

#' @describeIn thematic_with_theme obtain the current `theme`.
#' @param resolve whether or not `'auto'` values should be resolved before returning
#' @export
thematic_get_theme <- function(resolve = TRUE) {
  if (resolve) {
    auto_resolve_theme(.globals$theme)
  } else {
    .globals$theme
  }
}

#' @describeIn thematic_with_theme obtain a particular `theme` option (and provide a `default`
#'   if no `theme` is active).
#' @param name a theme element name (e.g., `fg`, `bg`, etc.)
#' @export
thematic_get_option <- function(name = "", default = NULL, resolve = TRUE) {
  if (length(name) != 1) {
    stop("`name` must be length 1", call. = FALSE)
  }
  theme <- thematic_get_theme(resolve = resolve)
  theme_names <- names(theme)
  if (length(theme) && length(setdiff(name, theme_names))) {
    stop(
      sprintf(
        "`name` must be one of the following: '%s'",
        paste(theme_names, collapse = "', '")
      ),
      call. = FALSE
    )
  }
  theme[[name]] %||% default
}

#' @describeIn thematic_with_theme obtain a mixture of the current `theme`'s `bg` and `fg`.
#' @param amounts value(s) between 0 and 1 specifying how much to mix `bg` (0) and `fg` (1).
#' @export
thematic_get_mixture <- function(amounts = 0.5, default = NULL) {
  if (!length(thematic_get_theme(resolve = FALSE))) {
    if (length(default)) {
      default <- rep_len(default, length(amounts))
    }
    return(default)
  }
  if (any(amounts < 0 | amounts > 1)) {
    stop("`amounts` must be between 0 and 1", call. = FALSE)
  }
  fg <- thematic_get_option("fg", resolve = TRUE)
  bg <- thematic_get_option("bg", resolve = TRUE)
  scales::colour_ramp(c(bg, fg))(amounts)
}

#' Font specification
#'
#' Specify a collection of font families. The first font family supported
#' by the relevant device (i.e., the device that is open, or will be opened, at
#' plotting time) is used by thematic. If a given font family is not supported
#' by the default, but is a [Google Font](https://fonts.google.com/) and
#' `install = TRUE`, the font will be downloaded, cached, and registered
#' for use the **showtext** and **ragg** packages.
#'
#' @param families a character vector of font families.
#' @param scale numerical constant applied to font sizes.
#' @param install whether to download and register font `families`
#' available via [Google Fonts](https://fonts.google.com/) (but unavailable to R).
#' After a successful download, fonts are cached (in a directory which
#' can be managed via [font_cache_set()]), and registered for use with
#' the **showtext** and **ragg** packages. If installation fails with
#' a valid internet connection, you may need to fetch the latest Google
#' Font information prior to installation (i.e., set `update = TRUE`).
#' @param update if `TRUE`, the latest Google Fonts are fetched and
#' any out-dated font cache is updated. Fetching the latest fonts requires
#' a Google Font API key (one is bundled with the package, but you can
#' set your own via an environment variable, `GFONT_KEY`).
#' @param quiet whether to suppress download messages.
#'
#' @return the input arguments as a list.
#' @seealso [thematic_save_plot()], [thematic_on()], [font_cache_set()]
#'
#' @export
font_spec <- function(families = "", scale = 1, install = is_installed("ragg") || is_installed("showtext"),
                      update = FALSE, quiet = TRUE) {

  if (update) {
    update_gfonts()
    update_gfonts_cache()
  }

  structure(
    list(
      families = tag_auto(families), scale = tag_auto(scale),
      install = install, quiet = quiet
    ),
    class = "font_spec"
  )
}

is_font_spec <- function(x) {
  inherits(x, "font_spec")
}

as_font_spec <- function(font) {
  if (is_font_spec(font)) return(font)
  if (isTRUE(is.na(font))) return(font_spec())
  if (identical(font, "auto")) return(font_spec(tag_auto(font), tag_auto(font)))
  if (is.character(font)) return(font_spec(font))

  stop("`font` must be either `NA`, a `font_spec()` object, or a character vector", call. = FALSE)
}


#' A color-blind safe qualitative colorscale (Okabe-Ito)
#'
#' This is the default `qualitative` colorscale in `thematic_on()`
#'
#' @param n number of colors.
#'
#' @return a vector of color codes.
#' @seealso [thematic_on()]
#' @references \url{https://jfly.uni-koeln.de/color/}
#' @export
okabe_ito <- function(n = NULL) {
  okabeIto <- c("#E69F00", "#009E73", "#0072B2", "#CC79A7", "#999999", "#D55E00", "#F0E442", "#56B4E9")
  if (is.null(n)) okabeIto else okabeIto[seq_len(n)]
}

#' Control parameters of the sequential colorscale
#'
#' Controls the default weighting and direction of the color gradient
#' derived from the `fg`, `bg`, and `accent` color (defined in `thematic_on()`).
#'
#' @param fg_weight a number (between 0 and 1) defining much of the `fg`
#' color should be mixed into the colorscale.
#' @param bg_weight a number (between 0 and 1) defining much of the `bg`
#' color should be mixed into the colorscale.
#' @param fg_low if `TRUE` (the default), the `fg` color is used for the
#' low end of the colorscale (rather than the high end).
#' @param n number of color codes.
#' @return a list of options for passing to the `sequential` argument of [thematic_on()].
#' @export
#' @examples
#'
#' # Gradient from fg to accent
#' fg <- sequential_gradient(1, 0)
#' thematic_on("black", "white", "salmon", sequential = fg)
#' ggplot2::qplot(1:10, 1:10, color = 1:10)
#'
#' # Gradient from accent -> bg
#' bg <- sequential_gradient(0, 1)
#' thematic_on("black", "white", "salmon", sequential = bg)
#' ggplot2::qplot(1:10, 1:10, color = 1:10)
#'
#' # Gradient from mix(accent, fg, 0.5) -> mix(accent, bg, 0.5)
#' mix <- sequential_gradient(0.5, 0.5)
#' thematic_on("black", "white", "salmon", sequential = mix)
#' ggplot2::qplot(1:10, 1:10, color = 1:10)
#'
#' # Use fg (instead of bg) for high end of scale
#' mix_flip <- sequential_gradient(0.5, 0.5, fg_low = FALSE)
#' thematic_on("black", "white", "salmon", sequential = mix_flip)
#' ggplot2::qplot(1:10, 1:10, color = 1:10)
#'
sequential_gradient <- function(fg_weight = 0.9, bg_weight = 0, fg_low = TRUE, n = 30) {
  if (any(fg_weight > 1 | fg_weight < 0)) {
    stop("`fg_weight` must be between 0 and 1.", call. = FALSE)
  }
  if (any(bg_weight > 1 | bg_weight < 0)) {
    stop("`bg_weight` must be between 0 and 1.", call. = FALSE)
  }
  if (n < 3) {
    stop("`n` must be 3 or more.", call. = FALSE)
  }
  if (!is.logical(fg_low)) {
    stop("`fg_low` must be `TRUE` or `FALSE`", call. = FALSE)
  }

  # Main idea: Interpolate between [fg+accent -> accent -> bg+accent]
  # For the endpoints the amount of blending of fg/bg and accent
  # depends on their perceptual distance
  function(fg, accent, bg, ...) {
    accent <- accent[1]
    if (anyNA(c(fg, accent, bg))) return(NA)

    fg_dist <- farver::compare_colour(
      farver::decode_colour(fg), farver::decode_colour(accent),
      from_space = "rgb", method = "cmc"
    )
    bg_dist <- farver::compare_colour(
      farver::decode_colour(bg), farver::decode_colour(accent),
      from_space = "rgb", method = "cmc"
    )
    total_dist <- bg_dist + fg_dist

    rng <- if (fg_low) {
      c(
        -fg_weight * as.numeric(fg_dist / total_dist),
        bg_weight * as.numeric(bg_dist / total_dist)
      )
    } else {
      c(
        -bg_weight * as.numeric(bg_dist / total_dist),
        fg_weight * as.numeric(fg_dist / total_dist)
      )
    }
    grid <- scales::rescale(
      seq(0, 1, length.out = n),
      to = pmax(pmin(rng + 0.5, 1), 0)
    )
    cols <- if (fg_low) c(fg, accent, bg) else c(bg, accent, fg)
    scales::colour_ramp(cols, alpha = TRUE)(grid)
  }
}

skip_if_not_installed("lattice")
library(lattice)
library(stats)

theme <-  thematic_theme(
  "black", "white", "pink",
  font = font_spec("Rock Salt", update = TRUE)
)

test_that("lattice baselines", {
  thematic_local_theme(theme)

  expect_doppelganger("dot_plot", dotplot(variety ~ yield | year * site, data=barley))

  expect_doppelganger("level_plot", {
      x <- seq(pi/4, 5 * pi, length.out = 100)
      y <- seq(pi/4, 5 * pi, length.out = 100)
      r <- as.vector(sqrt(outer(x^2, y^2, "+")))
      grid <- expand.grid(x=x, y=y)
      grid$z <- cos(r^2) * exp(-r/(pi^3))
      levelplot(z ~ x * y, grid, cuts = 50, scales=list(log="e"), xlab="",
                ylab="", main="Weird Function", sub="with log scales",
                colorkey = FALSE, region = TRUE)
    }
  )
  expect_doppelganger("density_plot", {
      histogram( ~ height | voice.part, data = singer,
                 xlab = "Height (inches)", type = "density",
                 panel = function(x, ...) {
                   panel.histogram(x, ...)
                   panel.mathdensity(dmath = dnorm, col = "black",
                                     args = list(mean=mean(x),sd=sd(x)))
                 } )
  }
  )

  expect_doppelganger("cleveland", {
      EE <- equal.count(ethanol$E, number=9, overlap=1/4)

      ## Constructing panel functions on the fly; prepanel
      xyplot(NOx ~ C | EE, data = ethanol,
             prepanel = function(x, y) prepanel.loess(x, y, span = 1),
             xlab = "Compression Ratio", ylab = "NOx (micrograms/J)",
             panel = function(x, y) {
               panel.grid(h = -1, v = 2)
               panel.xyplot(x, y)
               panel.loess(x, y, span=1)
             },
             aspect = "xy")
    })

  expect_doppelganger("xyplot", {
      xyplot(Sepal.Length + Sepal.Width ~ Petal.Length + Petal.Width | Species,
             data = iris, scales = "free", layout = c(2, 2),
             auto.key = list(x = .6, y = .7, corner = c(0, 0)))
    })

  expect_doppelganger("interaction_plot", {
      xyplot(decrease ~ treatment, OrchardSprays, groups = rowpos,
             type = "a",
             auto.key =
               list(space = "right", points = FALSE, lines = TRUE))
      })

  expect_doppelganger("bwplot", {
      bwplot(decrease ~ treatment, OrchardSprays, groups = rowpos,
             panel = "panel.superpose",
             panel.groups = "panel.linejoin",
             xlab = "treatment",
             key = list(lines = Rows(trellis.par.get("superpose.line"),
                                     c(1:7, 1)),
                        text = list(lab = as.character(unique(OrchardSprays$rowpos))),
                        columns = 4, title = "Row position"))
    })

  expect_doppelganger("settings", show.settings())


  thematic_local_theme(thematic_theme("black", "white", c("orange", "blue")))
  expect_doppelganger("settings2", show.settings())
})

skip_if_not_installed("ggplot2")

library(ggplot2)

test_that("works as expected with global ggthemes", {
  ggtheme <- theme_get()
  on.exit(theme_set(ggtheme), add = TRUE)
  theme_set(theme_void())
  thematic_on("black", "white", accent = NA, font = font_spec(scale = 1.5))
  expect_doppelganger("ggtheme-void", qplot(1:10))
  theme_set(theme_minimal())
  expect_doppelganger("ggtheme-minimal", qplot(1:10))
  thematic_off()
  expect_doppelganger("ggtheme-minimal-off", qplot(1:10))
})


test_that("works as expected with plot-specific themes", {
  thematic_on(bg = "#222222", fg = "white", accent = NA)
  on.exit(thematic_off(), add = TRUE)
  p <- ggplot(mtcars, aes(wt, mpg)) +
    geom_point() +
    facet_wrap(~cyl)
  p <- p + theme(strip.background = element_rect(fill = "purple"))
  expect_doppelganger("purple-strip", p)

  p <- qplot(1:10) + theme(
    axis.text = element_text(colour = "red", size = 5),
    axis.text.x = element_text(size = 50)
  )
  expect_doppelganger("axis-interitance", p)

  p <- qplot(1:10) + theme(
    text = element_text(colour = "red", size = 5),
    axis.text.x.bottom = element_text(size = 50)
  )
  expect_doppelganger("axis-interitance-distant", p)
  expect_doppelganger("axis-interitance-distant2", last_plot())
})


test_that("Global ggthemes are respected", {

  p <- ggplot(iris, aes(x = Sepal.Length, y = Sepal.Width, color =  Species)) +
    geom_point()

  with_ggtheme <- function(theme, expr) {
    old_theme <- theme_set(theme)
    on.exit(theme_set(old_theme), add = TRUE)
    thematic_with_theme(thematic_theme("black", "white", NA), expr)
  }

  # ggplot2 themes
  expect_doppelganger("bw", function() with_ggtheme(theme_bw(), p))
  expect_doppelganger("dark", function() with_ggtheme(theme_dark(), p))
  expect_doppelganger("minimal", function() with_ggtheme(theme_minimal(), p))
  expect_doppelganger("classic", function() with_ggtheme(theme_classic(), p))
  expect_doppelganger("light", function() with_ggtheme(theme_light(), p))
  expect_doppelganger("linedraw", function() with_ggtheme(theme_linedraw(), p))
  expect_doppelganger("void", function() with_ggtheme(theme_void(), p))

  # ggthemes
  skip_if_not_installed("ggthemes")
  ggthemes <- asNamespace("ggthemes")
  expect_doppelganger("wsj", function() with_ggtheme(ggthemes$theme_wsj(), p))
  expect_doppelganger("base", function() with_ggtheme(ggthemes$theme_base(), p))
  expect_doppelganger("solid", function() with_ggtheme(ggthemes$theme_solid(), p))
  expect_doppelganger("solarized", function() with_ggtheme(ggthemes$theme_solarized(), p))
  expect_doppelganger("economist", function() with_ggtheme(ggthemes$theme_economist(), p))
  expect_doppelganger("calc", function() with_ggtheme(ggthemes$theme_calc(), p))
  expect_doppelganger("fivethirtyeight", function() with_ggtheme(ggthemes$theme_fivethirtyeight(), p))
  expect_doppelganger("map", function() with_ggtheme(ggthemes$theme_map(), p))
  expect_doppelganger("excel", function() with_ggtheme(ggthemes$theme_excel(), p))
  expect_doppelganger("clean", function() with_ggtheme(ggthemes$theme_clean(), p))
  expect_doppelganger("excel_new", function() with_ggtheme(ggthemes$theme_excel_new(), p))
  expect_doppelganger("igray", function() with_ggtheme(ggthemes$theme_igray(), p))
  expect_doppelganger("foundation", function() with_ggtheme(ggthemes$theme_foundation(), p))
  expect_doppelganger("few", function() with_ggtheme(ggthemes$theme_few(), p))
  expect_doppelganger("pander", function() with_ggtheme(ggthemes$theme_pander(), p))
  expect_doppelganger("solarized_2", function() with_ggtheme(ggthemes$theme_solarized_2(), p))
  expect_doppelganger("gdocs", function() with_ggtheme(ggthemes$theme_gdocs(), p))
  expect_doppelganger("tufte", function() with_ggtheme(ggthemes$theme_tufte(), p))
  expect_doppelganger("par", function() with_ggtheme(ggthemes$theme_par(), p))
  expect_doppelganger("stata", function() with_ggtheme(ggthemes$theme_stata(), p))
  expect_doppelganger("hc", function() with_ggtheme(ggthemes$theme_hc(), p))
  skip_if_not_installed("dplyr")
  expect_doppelganger("economist_white", function() with_ggtheme(ggthemes$theme_economist_white(), p))

})

test_that("thematic_with_theme()", {
  expect_null(thematic_get_theme())
  expect_doppelganger("scatter-no-theme", function() { plot(1:10) })
  expect_doppelganger("scatter-with-theme", function() {
    thematic_with_theme(
      thematic_theme("black", "white", NA),
      plot(1:10)
    )
  })
  expect_doppelganger("scatter-no-theme2", function() { plot(1:10) })
})

skip_if_not_installed("ggplot2")

library(ggplot2)

test_that("Setting and getting the theme", {
  expect_null(thematic_get_theme())
  thematic_on("black", "white", "green")
  on.exit(thematic_off(), add = TRUE)
  x <- thematic_get_theme(resolve = FALSE)
  expect_true(is_thematic_theme(x))
  y <- thematic_on("darkblue", "skyblue", "purple")
  expect_identical(x, y)
  expect_doppelganger("purple", qplot(1:10, 1:10, color = 1:10))
  thematic_set_theme(y)
  expect_equal(thematic_get_theme(resolve = FALSE), y)
  expect_doppelganger("green", qplot(1:10, 1:10, color = 1:10))
  thematic_set_theme(NULL)
  expect_null(thematic_get_theme())
  expect_doppelganger("none", qplot(1:10, 1:10, color = 1:10))
})

test_that("Theme inheritance works", {
  thematic_on("black", "red", "green")
  on.exit(thematic_off(), add = TRUE)
  white <- thematic_theme(fg = "white", inherit = TRUE)
  thematic_on(fg = "white", inherit = TRUE)
  expect_equal(thematic_get_theme(resolve = FALSE), white)
  expect_equal(thematic_get_theme(), auto_resolve_theme(white))
  expect_doppelganger("white-green", qplot(1:10, 1:10, color = 1:10))
  thematic_on(accent = "red", inherit = TRUE)
  expect_doppelganger("white-red", qplot(1:10, 1:10, color = 1:10))
})

test_that("Getting options", {
  expect_null(thematic_get_theme())
  expect_equal(thematic_get_option("bg", "purple"), "purple")
  expect_equal(thematic_get_option("fg", "yellow"), "yellow")
  expect_equal(thematic_get_option("accent", "orange"), "orange")

  thematic_on("black", "white", "green")
  expect_equal(thematic_get_option("bg"), "black")
  expect_equal(thematic_get_option("fg"), "white")
  expect_equal(thematic_get_option("accent"), "green")
  expect_equal(
    thematic_get_mixture(seq(0, 1, by = 0.1)),
    c('#000000', '#1B1B1B', '#303030', '#474747', '#5E5E5E', '#777777', '#919191', '#ABABAB', '#C6C6C6', '#E2E2E2', '#FFFFFF')
  )
  thematic_off()
})

skip_if_not_installed("ggplot2")

test_that("Sequential gradients works as expected", {
  library(ggplot2)

  on.exit(thematic_off(), add = TRUE)

  # Gradient from fg to accent
  fg <- sequential_gradient(1, 0)
  thematic_on("black", "white", "salmon", sequential = fg)
  p <- ggplot2::qplot(1:10, 1:10, color = 1:10)
  expect_doppelganger("fg-accent", p)

  # Gradient from accent -> bg
  bg <- sequential_gradient(0, 1)
  thematic_on("black", "white", "salmon", sequential = bg)
  p <- ggplot2::qplot(1:10, 1:10, color = 1:10)
  expect_doppelganger("accent-bg", p)

  # Gradient from mix(accent, fg, 0.5) -> mix(accent, bg, 0.5)
  mix <- sequential_gradient(0.5, 0.5)
  thematic_on("black", "white", "salmon", sequential = mix)
  p <- ggplot2::qplot(1:10, 1:10, color = 1:10)
  expect_doppelganger("half-half", p)

  # Use fg (instead of bg) for high end of scale
  mix_flip <- sequential_gradient(0.5, 0.5, fg_low = FALSE)
  thematic_on("black", "white", "salmon", sequential = mix_flip)
  ggplot2::qplot(1:10, 1:10, color = 1:10)
  expect_doppelganger("half-half-reversed", p)
})

skip_if_not_installed("ggplot2")
library(ggplot2)

theme <- thematic_theme(
  "#444444", "#e4e4e4", "#749886",
  font = font_spec("Oxanium", scale = 1.25, update = TRUE)
)

test_that("ggplot baselines", {
  thematic_local_theme(theme)

  ids <- factor(c("1.1", "2.1", "1.2", "2.2", "1.3", "2.3"))
  values <- data.frame(
    id = ids,
    value = c(3, 3.1, 3.1, 3.2, 3.15, 3.5)
  )
  positions <- data.frame(
    id = rep(ids, each = 4),
    x = c(2, 1, 1.1, 2.2, 1, 0, 0.3, 1.1, 2.2, 1.1, 1.2, 2.5, 1.1, 0.3,
          0.5, 1.2, 2.5, 1.2, 1.3, 2.7, 1.2, 0.5, 0.6, 1.3),
    y = c(-0.5, 0, 1, 0.5, 0, 0.5, 1.5, 1, 0.5, 1, 2.1, 1.7, 1, 1.5,
          2.2, 2.1, 1.7, 2.1, 3.2, 2.8, 2.1, 2.2, 3.3, 3.2)
  )
  huron <- data.frame(year = 1875:1972, level = as.vector(LakeHuron))

  dsample <- diamonds[sample(nrow(diamonds), 1000), ]


  expect_doppelganger("GeomPointColor", ggplot(mtcars, aes(wt, mpg, color = cyl)) + geom_point())
  expect_doppelganger("GeomAbline", {
    ggplot(mtcars, aes(mpg, wt)) +
      geom_point() +
      geom_hline(aes(yintercept = wt), data.frame(cyl = c(4, 6, 8), wt = c(2.28, 3.11, 4.00))) +
      facet_wrap(~ cyl)
  })
  expect_doppelganger("GeomBar", ggplot(mpg, aes(class)) + geom_bar(aes(weight = displ)))
  expect_warning(expect_doppelganger("GeomBin2d", ggplot(diamonds, aes(x, y)) + xlim(4, 10) + ylim(4, 10) + geom_bin2d()))
  expect_doppelganger("GeomBox", ggplot(mpg, aes(class, hwy)) + geom_boxplot())
  # For some reason these give slightly different svgs on different platforms?
  #expect_doppelganger("GeomContour", ggplot(faithfuld, aes(waiting, eruptions, z = density)) + geom_contour())
  #expect_doppelganger("GeomContourRaster", ggplot(faithfuld, aes(waiting, eruptions, z = density)) + geom_raster(aes(fill = density)) + geom_contour())
  expect_doppelganger("GeomCount", ggplot(mpg, aes(cty, hwy)) + geom_count())
  expect_doppelganger("GeomDensity", ggplot(diamonds, aes(carat)) + geom_density())
  expect_doppelganger("GeomDensityColor", ggplot(diamonds, aes(depth, fill = cut, color = cut)) + geom_density(alpha = 0.1))
  expect_doppelganger("GeomDotPlot",  ggplot(mtcars, aes(x = mpg)) + geom_dotplot())
  expect_doppelganger(
    "GeomError", {
      ggplot(data.frame(
        trt = factor(c(1, 1, 2, 2)),
        resp = c(1, 5, 3, 4),
        group = factor(c(1, 2, 1, 2)),
        se = c(0.1, 0.3, 0.3, 0.2)
      ), aes(resp, trt, color = group)) + geom_point() +
        geom_errorbarh(aes(xmax = resp + se, xmin = resp - se))
    })
  expect_doppelganger("GeomHistogram",  ggplot(diamonds, aes(price, fill = cut)) + geom_histogram(binwidth = 500))
  expect_doppelganger("GeomJitter", ggplot(mpg, aes(cyl, hwy)) + geom_jitter(width = 0.25))
  expect_doppelganger(
    "GeomErrorbar",  {
      ggplot(data.frame(
        trt = factor(c(1, 1, 2, 2)),
        resp = c(1, 5, 3, 4),
        group = factor(c(1, 2, 1, 2)),
        upper = c(1.1, 5.3, 3.3, 4.2),
        lower = c(0.8, 4.6, 2.4, 3.6)
      ), aes(trt, resp, fill = group)) +
        geom_col(position = "dodge") +
        geom_errorbar(aes(ymin = lower, ymax = upper), position = "dodge", width = 0.25)
    })
  expect_doppelganger(
    "GeomPolygon", {
      ggplot(values) +
        geom_map(aes(map_id = id), map = positions) +
        expand_limits(positions)
    }
  )
  expect_doppelganger(
    "GeomPolygon2", {
      ggplot(values, aes(fill = value)) +
        geom_map(aes(map_id = id), map = positions) +
        expand_limits(positions)
    }
  )
  expect_doppelganger("GeomLine", ggplot(economics_long, aes(date, value01, group = variable)) + geom_line())
  expect_doppelganger("GeomLine2", ggplot(economics_long, aes(date, value01, color = variable)) + geom_line())
  expect_doppelganger("GeomPoint", ggplot(dsample, aes(carat, price)) + geom_point(alpha = 0.1))
  expect_doppelganger("GeomPoint2", ggplot(mtcars, aes(wt, mpg, color = factor(cyl), shape = factor(cyl))) + geom_point())
  expect_doppelganger("GeomRibbon", {
    ggplot(huron, aes(year)) +
      geom_ribbon(aes(ymin = level - 1, ymax = level + 1), fill = "grey70") +
      geom_line(aes(y = level))
  })
  expect_doppelganger("GeomRug", ggplot(mtcars, aes(wt, mpg)) + geom_point() + geom_rug())
  expect_doppelganger("GeomRug2", ggplot(mtcars, aes(wt, mpg)) + geom_point() + geom_rug(outside = TRUE) + coord_cartesian(clip = "off"))
  expect_doppelganger(
    "GeomCurve", {
      df <- data.frame(x1 = 2.62, x2 = 3.57, y1 = 21.0, y2 = 15.0)
      ggplot(mtcars, aes(wt, mpg)) +
        geom_point() +
        geom_curve(aes(x = x1, y = y1, xend = x2, yend = y2, color = "curve"), data = df) +
        geom_segment(aes(x = x1, y = y1, xend = x2, yend = y2, color = "segment"), data = df) +
        geom_curve(aes(x = x1, y = y1, xend = x2, yend = y2), data = df, curvature = -0.2)
    }
  )

  expect_doppelganger(
    "GeomSmooth", {
      ggplot(dsample, aes(carat, price)) +
        geom_point(alpha = 0.2) +
        geom_smooth() +
        facet_wrap(~cut) + ggtitle("Diamond price by carat and cut")
    })
  expect_doppelganger(
    "GeomSmooth2", {
      ggplot(mpg, aes(displ, hwy, color = class)) +
        geom_point() +
        geom_smooth(se = FALSE, method = lm)
    }
  )
  expect_doppelganger(
    "GeomSpoke", {
      df <- expand.grid(x = 1:10, y=1:10)
      df$angle <- runif(100, 0, 2*pi)
      df$speed <- runif(100, 0, sqrt(0.1 * df$x))

      ggplot(df, aes(x, y)) +
        geom_point() +
        geom_spoke(aes(angle = angle), radius = 0.5)
    }
  )
  expect_doppelganger("GeomText", ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars))) + geom_text())
  expect_doppelganger("GeomLabel", ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars))) + geom_label())
  expect_doppelganger("GeomTile", {
    ggplot(data.frame(
      x = rep(c(2, 5, 7, 9, 12), 2),
      y = rep(c(1, 2), each = 5),
      z = factor(rep(1:5, each = 2)),
      w = rep(diff(c(0, 4, 6, 8, 10, 14)), 2)
    ), aes(x, y, width = w)) + geom_tile(aes(fill = z))
  })
  expect_doppelganger("GeomViolin", ggplot(mtcars, aes(factor(cyl), mpg)) + geom_violin())
})

test_that("Scale defaults can be overridden", {
  thematic_local_theme(theme)

  expect_doppelganger("sequential-color", {
    ggplot(mtcars, aes(wt, mpg, color = cyl)) +
      geom_point() +
      scale_color_gradient2(midpoint = 6)
  })

  expect_doppelganger("sequential-fill", {
    ggplot(faithfuld, aes(waiting, eruptions, z = density)) +
      geom_raster(aes(fill = density)) +
      scale_fill_gradient2(midpoint = 0.02)
  })

  expect_doppelganger("qualitative-color", {
    ggplot(economics_long) +
      geom_line(aes(date, value01, color = variable)) +
      scale_color_brewer(palette = "Accent")
  })

  expect_doppelganger("qualitative-fill", {
    ggplot(mtcars, aes(mpg, fill = factor(am))) +
      geom_density() +
      scale_fill_brewer(palette = "Accent")
  })

})

test_that("sf integration", {
  skip_if_not_installed("sf")
  skip_if_not_installed("rgdal")
  skip_if_not(Sys.info()[["sysname"]] == "Darwin")
  st_read <- getFromNamespace("st_read", "sf")
  st_transform <- getFromNamespace("st_transform", "sf")
  st_centroid <- getFromNamespace("st_centroid", "sf")

  thematic_local_theme(theme)

  nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
  nc_3857 <- st_transform(nc, 3857)
  nc_3857$mid <- st_centroid(nc_3857$geometry)

  p <- ggplot(nc_3857) +
    geom_sf(color = "white") +
    geom_sf(aes(geometry = mid, size = AREA), show.legend = "point")

  expect_doppelganger("GeomSf", p)
})

theme2 <- thematic_theme(
  bg = "black", fg = "white", accent = "salmon",
  font_spec("Oxanium", scale = 1.25)
)

test_that("gridExtra integration", {
  skip_if_not_installed("gridExtra")

  thematic_local_theme(theme2)
  p1 <- qplot(x = 1:10, y = 1:10, color = 1:10)
  p2 <- qplot(x = 1:10, y = 1:10, color = 1:10)
  expect_doppelganger(
    "grid-arrange", getFromNamespace("grid.arrange", "gridExtra")(p1, p2)
  )
})

test_that("patchwork integration", {
  skip_if_not_installed("patchwork")
  ggplot2 <- "patchwork"
  library(ggplot2)
  p1 <- qplot(x = 1:10, y = 1:10, color = 1:10)
  p2 <- qplot(x = 1:10, y = 1:10, color = 1:10)
  thematic_local_theme(theme2)
  expect_doppelganger("patchwork", p1 + p2)
})

# https://github.com/rstudio/thematic/pull/83
test_that("element_blank() inherits properly", {
  thematic_local_theme(theme2)
  expect_doppelganger(
    "no-y-text",
    qplot(x = 1:10, y = 1:10) + theme(axis.text.y = element_blank())
  )
})


test_that("Relevant ggplot_build() method(s) are owned by thematic", {
  ggplot_uses_S7 <- "class_ggplot" %in% getNamespaceExports("ggplot2")
  skip_if(ggplot_uses_S7)
  thematic_local_theme(theme2)
  expect_equal(
    body(getFromNamespace("ggplot_build.ggplot", "ggplot2")),
    body(ggthematic_build)
  )
  skip_if_not_installed("gganimate")
  expect_equal(
    body(getFromNamespace("ggplot_build.gganim", "gganimate")),
    body(ggthematic_build)
  )
})

test_that("base baselines", {
  thematic_local_theme(
    thematic_theme("black", "white", "violet", font_spec("Amatic SC", 1.5, update = TRUE))
  )

  expect_doppelganger("scatter", function() { plot(1:10) })
  expect_doppelganger("scatter-cols", function() { plot(1:10, col = 1:10) })
  expect_doppelganger("lines", function() {
    plot(1:10, (1:10)^2, type = "b", pch = 19, xlab = "x", ylab = "y")
  })
  expect_doppelganger("lines-accent", function() {
    plot(1:10, (1:10)^2, type = "b", pch = 19, xlab = "x", ylab = "y", col = thematic_get_option("accent"))
  })
  expect_doppelganger("lines-cols", function() {
    plot(1:10, (1:10)^2, type = "b", pch = 19, xlab = "x", ylab = "y", col = 1:10)
  })
  expect_doppelganger("smooth-accent", function() {
    # draw a smooth line through a scatter plot
    plot(cars, main = "Stopping Distance versus Speed")
    lines(stats::lowess(cars), col = thematic_get_option("accent"))
  })
  expect_doppelganger("hist", function() { hist(rnorm(100)) })
  expect_doppelganger("barplot", function() {
    barplot(VADeaths[1:3, "Rural Male"])
  })
  expect_doppelganger("barplot-accent", function() {
    barplot(VADeaths[1:3, "Rural Male"], col = thematic_get_option("accent"))
  })
  expect_doppelganger("stripchart", function() {
    stripchart(len ~ dose, data = ToothGrowth,
               pch = 19, frame = FALSE, vertical = TRUE,
               method = "jitter")
  })
  expect_doppelganger("stripchart-cols", function() {
    stripchart(len ~ dose, data = ToothGrowth,
               pch = 19, col = thematic_get_option("qualitative"),
               method = "jitter")
  })

  expect_doppelganger("pies", function() {
    df <- data.frame(
      group = c("Male", "Female", "Child"),
      value = c(25, 25, 50)
    )
    par(mfrow = c(2, 1))
    pie(df$value, labels = df$group, radius = 1)
    pie(df$value, labels = df$group, radius = 1, col = thematic_get_option("qualitative"))
  })

  expect_doppelganger("qq", function() {
    stats::qqnorm(ToothGrowth$len, pch = 1, frame = FALSE)
    stats::qqline(ToothGrowth$len, col = thematic_get_option("accent"), lwd = 2)
  })

  expect_doppelganger(
    "pairs",
    function() {
      pairs(
        iris[,1:4], pch = 19, cex = 0.5,
        col = thematic_get_option("qualitative")[iris$Species],
        lower.panel = NULL
      )
    }
  )

})

set_hooks <- function() {
  setHook("before.plot.new", base_before_hook)
  setHook("before.grid.newpage", grid_before_hook)
  setHook("plot.new", base_plot_hook)
  setHook("grid.newpage", grid_plot_hook)
}

remove_hooks <- function() {
  remove_hook("before.plot.new", base_before_hook)
  remove_hook("before.grid.newpage", grid_before_hook)
  remove_hook("plot.new", base_plot_hook)
  remove_hook("grid.newpage", grid_plot_hook)
}

remove_hook <- function(name, hook) {
  hooks <- getHook(name)
  is_thematic <- vapply(hooks, function(x) identical(x, hook), logical(1))
  setHook(name, hooks[!is_thematic], "replace")
}

base_before_hook <- function() {
  .globals$theme <- auto_resolve_theme(.globals$theme)
  # populates .globals$theme$font$family based on the first families we can support
  resolve_font_family(type = "base")
  # update graphical parameters
  base_params_set()
  base_palette_set()
}

#' @include globals.R
.globals$recordShowtext <- FALSE
base_plot_hook <- function() {
  # If showtext_begin() was called before.plot.new (this timing seems important for showtext to be effective)
  # then we record that call in plot.new hook (this timing seems important for recordGraphics to be effective).
  # Note that by recording this call, showtext works more generally in RStudio and Shiny
  # https://github.com/yixuan/showtext/issues/41#issuecomment-670983568
  if (isTRUE(.globals$recordShowtext)) {
    grDevices::recordGraphics(showtext::showtext_begin(), list(), getNamespace("showtext"))
  }
}

grid_before_hook <- function() {
  .globals$theme <- auto_resolve_theme(.globals$theme)
  # populates .globals$theme$font$family based on the first families we can support
  resolve_font_family(type = "grid")
  # update ggplot2/lattice defaults
  ggplot_build_set()
  lattice_print_set()
}

grid_plot_hook <- base_plot_hook


resolve_font_family <- function(type = c("base", "grid")) {
  font <- .globals$theme$font
  families <- font$families
  is_default_font <- identical(families, "")

  # Return early if default font family
  if (is_default_font) return()

  # Returns the name of the currently active device
  # (and, if none is active, the name of the one that *will be* used)
  dev_name <- infer_device()

  # Make sure fig.showtext = TRUE in knitr (if this is a non-ragg device)
  # (We set this .onLoad, but it only applies for the _next_ chunk)
  if (isTRUE(getOption("knitr.in.progress"))) {
    dev <- knitr::opts_current$get("dev")
    show <- knitr::opts_current$get("fig.showtext")
    if (!identical(show, TRUE) && !identical(dev, "ragg_png")) {
      stop("The fig.showtext code chunk option must be TRUE", call. = FALSE)
    }
  }

  if (is_installed("showtext") && !is_ragg_device(dev_name)) {
    showtext::showtext_begin()
    .globals$recordShowtext <- TRUE
  }

  # Since can_render() needs to open the device to detect font support,
  # we need to be able to map a device name to a function
  dev_fun <- get_device_function(dev_name)
  if (!is.function(dev_fun)) {
    stop("Internal error: get_device_function() should return a function.", call. = FALSE)
  }

  type <- match.arg(type)
  for (i in seq_along(families)) {
    family <- families[[i]]

    # Bootstrap 4 defaults to a CSS system font family...
    # we know we can't render those, so warn and try the next
    if (family %in% generic_css_families()) {
      maybe_warn(
        "Generic CSS font families (e.g. '", family, "') aren't supported. ",
        "Consider using a Google Font family instead https://fonts.google.com/",
        id = paste0("generic-css-family-", family)
      )
      next
    }

    # Register Google Fonts cache
    try_register_gfont_cache(family, systemfonts = is_ragg_device(dev_name))
    # If we can already render the font family, do no more!
    if (can_render(family, type, dev_fun, dev_name)) {
      break
    }
    # Download/register and try again
    if (font$install) {
      try_gfont_download_and_register(family, font$quiet, systemfonts = is_ragg_device(dev_name))
    }
    if (can_render(family, type, dev_fun, dev_name)) {
      break
    }

    # There's two scenarios that could cause a font to fail related to availability of a rendering device.
    pkg_install_msg <- if (!is_installed("ragg") && !is_installed("showtext")) {
      "Install ragg and/or showtext to render custom fonts. "
    } else if (is_installed("ragg") & !is_ragg_device(dev_name)) {
      "To render custom fonts, either use a ragg device or install the showtext package"
    } else {
      ""
    }

    if (nzchar(pkg_install_msg)) {
      # If we have a problem related to loaded packages rendering the font then
      # we want to just use the default font. Otherwise the plot when attempting
      # to use a non-existent font and will crash with a bunch of error messages
      # that bury our useful one above.
      family <- ""
    }

    # Try our best to give informative warning
    maybe_warn(
      "It seems the current graphics device '", dev_name, "' ",
      "is unable to render the requested font family '", family, "'. ",
      pkg_install_msg,
      id = "cant-render-font"
    )
  }

  set_font_family(family)
}

set_font_family <- function(family) {
  .globals$theme$font$family <- family
  family
}


can_render <- function(family, type = c("base", "grid"), dev_fun, dev_name) {
  # ragg devices don't produce a warning if the font family is not
  # available; instead, it uses whatever match_font(family) gives.
  # Therefore, it seems reasonable to assume the font is available
  # if the match resolves to something other than a generic font family
  if (is_ragg_device(dev_name)) {
    if (family %in% c("sans", "serif", "mono", "emoji")) return(TRUE)
    f <- systemfonts::match_fonts(family)
    is_available <- !identical(f, systemfonts::match_fonts("sans")) &&
      !identical(f, systemfonts::match_fonts("serif")) &&
      !identical(f, systemfonts::match_fonts("mono")) &&
      !identical(f, systemfonts::match_fonts("emoji"))
    return(is_available)
  }

  # temporarily disable thematics plot hooks
  # (otherwise, we'd get caught in an infinite loop)
  remove_hooks()
  on.exit(set_hooks(), add = TRUE)

  # To see if we have font support on the given device without
  # generating multiple pages, (temporarily) open the device
  # to render some text
  attempt_with_device(
    dev_fun = dev_fun,
    tryCatch(
      {
        if (is_installed("showtext")) showtext::showtext_begin()
        if (type == "grid") {
          grid.newpage()
          grid.text("testing", x = 0.5, y = 0.5, gp = gpar(fontfamily = family))
        } else {
          plot(1, family = family)
        }
        TRUE
      },
      warning = function(w) {
        !grepl(
          family,
          paste(w$message, collapse = "\n"),
          fixed = TRUE
        )
      },
      error = function(e) { FALSE }
    )
  )
}


# Do our best to map the name of the current device to an
# actual device function
get_device_function <- function(name = infer_device()) {

  # first, resolve known cases where the .Device name
  # doesn't quite map to the relevant function name
  if ("win.metafile:" == name) {
    return(getFromNamespace("win.metafile", "grDevices"))
  }

  # Note that quartz defaults to an on-screen device,
  # so this needs to set to an off-screen type
  if ("quartz_off_screen" == name) {
    return(grDevices::png)
  }

  if ("RStudioGD" == name) {
    if (!rstudioapi::isAvailable("1.4")) {
      return(grDevices::png)
    }
    # RStudio 1.4 introduced a configurable graphics backend https://github.com/rstudio/rstudio/pull/6520
    backend <- tryNULL(readRStudioPreference("graphics_backend"))
    if ("ragg" %in% backend && is_installed("ragg")) {
      return(ragg::agg_png)
    }
    if ("windows" %in% backend && Sys.info()[["sysname"]] == "Windows") {
      return(utils::getFromNamespace("win.metafile", "grDevices"))
    }
    return(grDevices::png)
  }

  # Effectively what dev.new() does to find a device function from a string
  # https://github.com/wch/r-source/blob/d6c208e4/src/library/grDevices/R/device.R#L291-L293
  if (exists(name, .GlobalEnv)) {
    return(get(name, .GlobalEnv))
  }
  if (exists(name, asNamespace("grDevices"))) {
    return(get(name, asNamespace("grDevices")))
  }
  # Other important (non-grDevices) devices
  switch(
    name,
    agg_png = ragg::agg_png,
    agg_tiff = ragg::agg_tiff,
    agg_ppm = ragg::agg_ppm,
    agg_jpeg = ragg::agg_jpeg,
    Cairo = Cairo::Cairo,
    devSVG = svglite::svglite,
    devSVG_vdiffr = svglite::svglite,
    dsvg_device = ggiraph::dsvg,
    httpgd = getFromNamespace("hgd", "httpgd"),
    # https://github.com/posit-dev/positron/issues/2919#issuecomment-2219653110
    "Positron Graphics Device" = grDevices::png,
    # TODO: support cairoDevices? tikz?
    stop(
      "thematic doesn't (yet) support the '", name, "' graphics device. ",
      "Please report this error to https://github.com/rstudio/thematic/issues/new",
      call. = FALSE
    )
  )
}

is_ragg_device <- function(dev_name) {
  dev_name %in% paste0("agg_", c("png", "tiff", "ppm", "jpeg"))
}



# https://drafts.csswg.org/css-fonts-4/#generic-font-families
generic_css_families <- function() {
  c(
    "serif", "sans-serif", "cursive", "fantasy", "monospace",
    "system-ui", "emoji", "math", "fangsong",
    "ui-serif", "ui-sans-serif", "ui-monospace", "ui-rounded",
    # not part of the official spec (earlier versions of system-ui)
    "-apple-system" , "BlinkMacSystemFont"
  )
}

skip_on_cran()
skip_if_not_installed("shinytest2")
skip_if_not_installed("callr")
skip_if_not_installed("ggplot2")
skip_if_not_installed("lattice")


test_app <- function(appDir) {
  getFromNamespace("test_app", "shinytest2")(appDir, check_setup = FALSE)
}

platform <- getFromNamespace("platform_variant", "shinytest2")(
  r_version = FALSE
)


# Default to running tests on macOS
do_tests <- identical(platform, "mac")
skip_if_not(as.logical(Sys.getenv("SHINYTEST_RUN_TESTS", do_tests)))


test_that("Custom fonts with ragg", {
  skip_if_not_installed("ragg")
  test_app("agg_png")
})

test_that("runtime: shiny auto-theming", {
  skip_if_not(identical("win", platform))
  test_app("shiny_runtime")
})

# The remaining tests rely on showtext to render custom fonts
skip_if_not_installed("showtext")

test_that("Auto theming in shiny works", {
  # These shiny apps currently use remote urls to import Google fonts
  skip_if_offline()
  skip_if_not(capabilities()[["aqua"]])
  test_app("auto_theme_shiny/base")
  test_app("auto_theme_shiny/lattice")
  test_app("auto_theme_shiny/ggplot2")
  test_app("auto_theme_shiny/local_theme")
  test_app("auto_theme_shiny/with_theme")
})

test_that("Custom fonts with Cairo package", {
  skip_if_not_installed("Cairo")
  test_app("CairoPNG")
})

test_that("Custom fonts with quartz device", {
  skip_if_not(capabilities()[["aqua"]])
  test_app("quartz_png")
})

test_that("Custom fonts with cairo capabilities", {
  skip_if_not(capabilities()[["cairo"]])
  test_app("cairo_png")
  # TODO: figure out why this is failing on mac
  skip_if_not(!identical(platform, "mac"))
  test_app("cairo_svg")
})

# Remaining tests are rmarkdown specific
skip_if_not_installed("rmarkdown")

test_that("Auto theming in rmarkdown works", {
  skip_if_not(capabilities()[["aqua"]])
  test_app("auto_theme_rmd/darkly_rmd")
})

test_that("Can render non-custom fonts in rmarkdown with quartz png", {
  skip_if_not(capabilities()[["aqua"]])
  test_app("quartz_png_rmd")
})

test_that("Can render non-custom fonts in rmarkdown with CairoPNG", {
  skip_if_not_installed("Cairo")
  test_app("CairoPNG_rmd")
})

test_that("pdf_document compiles without error", {
  skip_if_not(!identical("win", platform))


  outfile <- rmarkdown::render("pdf.Rmd", quiet = TRUE)
  expect_true(file.exists(outfile))
  unlink(outfile)
})

---
runtime: shiny
output: 
  html_document:
    theme: darkly
---

```{r setup, include=FALSE}
library(thematic)
thematic_shiny()
knitr::opts_chunk$set(echo = FALSE)
```

<style>
  body{color: red}
</style>

## ggplot

```{r}
renderPlot({
  ggplot2::qplot(1:10)
})
```

---
title: "Untitled"
output: 
  html_document_base:
    theme: 
      bootswatch: {{theme}}
    self_contained: false
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  echo = FALSE,
  fig.path = "./"
)
thematic::thematic_rmd()
```

```{r ggplot}
library(ggplot2)
qplot(x = 1:10, y = 1:10, color = 1:10)
```

```{r lattice}
lattice::show.settings()
```

```{r base}
image(volcano, col = thematic_get_option("sequential"))
```

---
title: ""
output: 
  html_document_base:
    theme: null
    self_contained: false
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  echo = FALSE,
  fig.path = "./",
  dev = {{device}}
)
library(thematic)
thematic_rmd(
  "#444444", "#e4e4e4", "#749886", 
  font = font_spec("Liu Jian Mao Cao", scale = 1.25, update = TRUE)
)
```

```{r ggplot}
library(ggplot2)
qplot(x = 1:10, y = 1:10, color = 1:10)
```

```{r lattice}
lattice::show.settings()
```

```{r base}
image(volcano, col = thematic_get_option("sequential"))
```

---
output: pdf_document
---

```{r}
library(thematic)
thematic_rmd(
  "#444444", "#e4e4e4", "#749886", 
  font = font_spec("Liu Jian Mao Cao", scale = 1.25, update = TRUE)
)
```

```{r ggplot}
library(ggplot2)
qplot(x = 1:10, y = 1:10, color = 1:10)
```

```{r lattice}
lattice::show.settings()
```

```{r base}
image(volcano, col = thematic_get_option("sequential", resolve = FALSE))
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thematic.R
\name{font_spec}
\alias{font_spec}
\title{Font specification}
\usage{
font_spec(
  families = "",
  scale = 1,
  install = is_installed("ragg") || is_installed("showtext"),
  update = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{families}{a character vector of font families.}

\item{scale}{numerical constant applied to font sizes.}

\item{install}{whether to download and register font \code{families}
available via \href{https://fonts.google.com/}{Google Fonts} (but unavailable to R).
After a successful download, fonts are cached (in a directory which
can be managed via \code{\link[=font_cache_set]{font_cache_set()}}), and registered for use with
the \strong{showtext} and \strong{ragg} packages. If installation fails with
a valid internet connection, you may need to fetch the latest Google
Font information prior to installation (i.e., set \code{update = TRUE}).}

\item{update}{if \code{TRUE}, the latest Google Fonts are fetched and
any out-dated font cache is updated. Fetching the latest fonts requires
a Google Font API key (one is bundled with the package, but you can
set your own via an environment variable, \code{GFONT_KEY}).}

\item{quiet}{whether to suppress download messages.}
}
\value{
the input arguments as a list.
}
\description{
Specify a collection of font families. The first font family supported
by the relevant device (i.e., the device that is open, or will be opened, at
plotting time) is used by thematic. If a given font family is not supported
by the default, but is a \href{https://fonts.google.com/}{Google Font} and
\code{install = TRUE}, the font will be downloaded, cached, and registered
for use the \strong{showtext} and \strong{ragg} packages.
}
\seealso{
\code{\link[=thematic_save_plot]{thematic_save_plot()}}, \code{\link[=thematic_on]{thematic_on()}}, \code{\link[=font_cache_set]{font_cache_set()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thematic.R
\name{thematic_on}
\alias{thematic_on}
\alias{thematic_off}
\alias{thematic_theme}
\alias{thematic_shiny}
\alias{thematic_rmd}
\title{Enable (or disable) simplified theming of R graphics.}
\usage{
thematic_on(
  bg = "auto",
  fg = "auto",
  accent = "auto",
  font = NA,
  sequential = sequential_gradient(),
  qualitative = okabe_ito(),
  inherit = FALSE
)

thematic_off()

thematic_theme(
  bg = "auto",
  fg = "auto",
  accent = "auto",
  font = NA,
  sequential = sequential_gradient(),
  qualitative = okabe_ito(),
  inherit = FALSE
)

thematic_shiny(
  bg = "auto",
  fg = "auto",
  accent = "auto",
  font = NA,
  sequential = sequential_gradient(),
  qualitative = okabe_ito(),
  inherit = FALSE,
  session = shiny::getDefaultReactiveDomain()
)

thematic_rmd(
  bg = "auto",
  fg = "auto",
  accent = "auto",
  font = NA,
  sequential = sequential_gradient(),
  qualitative = okabe_ito(),
  inherit = FALSE
)
}
\arguments{
\item{bg}{a background color.}

\item{fg}{a foreground color.}

\item{accent}{a color for making certain graphical markers 'stand out'
(e.g., the fitted line color for \code{\link[ggplot2:geom_smooth]{ggplot2::geom_smooth()}}).
Can be 2 colors for lattice (stroke vs fill accent).}

\item{font}{a \code{\link[=font_spec]{font_spec()}} object. If missing, font defaults are not altered.}

\item{sequential}{a color palette for graphical markers that encode
numeric values. Can be a vector of color codes or a
\code{\link[=sequential_gradient]{sequential_gradient()}} object.}

\item{qualitative}{a color palette for graphical markers that encode
qualitative values (won't be used in ggplot2 when the number of data
levels exceeds the max allowed colors). Defaults to \code{\link[=okabe_ito]{okabe_ito()}}.}

\item{inherit}{should non-specified values inherit from the previous theme?}

\item{session}{see \code{\link[shiny:onStop]{shiny::onStop()}}.}
}
\value{
\code{\link[=thematic_theme]{thematic_theme()}} returns a theme object as a list (which can be
activated with \code{\link[=thematic_with_theme]{thematic_with_theme()}} or \code{\link[=thematic_set_theme]{thematic_set_theme()}}).

\code{\link[=thematic_on]{thematic_on()}}, \code{\link[=thematic_off]{thematic_off()}}, and \code{\link[=thematic_shiny]{thematic_shiny()}} all return
the previous global theme.
}
\description{
A unified interface for theming \strong{ggplot2}, \strong{base}, and \strong{lattice} graphics
based on a handful of styling options. In some cases (most notably in a \strong{shiny} runtime),
these options can automatically resolve to relevant CSS styles (see the "Auto theming"
section below).
}
\section{Auto theming}{


The \code{bg}, \code{fg}, \code{accent}, and \code{font} arguments all support a value of \code{'auto'},
which are all resolved, at plot time, based on the execution environment. In a
\strong{shiny} runtime, resolution of auto values should always work as expect; but
in other contexts, auto values may lead to wrong or surprising results. In that
case, auto resolution logic can be customized (see \code{\link[=auto_config_set]{auto_config_set()}} for more details).
}

\section{Global vs. local theming}{


\code{thematic_on()} enables thematic in a global fashion (that is, it impacts all
future plots, up until \code{thematic_off()} is called). To use thematic in local fashion,
first create a theme with \code{\link[=thematic_theme]{thematic_theme()}}, then provide it to \code{\link[=thematic_with_theme]{thematic_with_theme()}}
(or similar). To use thematic in a global fashion up until a \strong{shiny}
app exits, use \code{thematic_shiny()} (which cleans up after itself once the next shiny
app that exits using \code{\link[shiny:onStop]{shiny::onStop()}}). To use thematic in a global fashion up until
a \strong{rmarkdown} document finishes rendering, use \code{thematic_rmd()}.
}

\section{Color values}{


Colors (e.g., \code{bg}, \code{fg}, \code{accent}) may be any value understood by \code{\link[=col2rgb]{col2rgb()}}
or \code{htmltools::parseCssColors()} (i.e., may be any valid R or CSS color string).
}

\examples{
# simple dark mode
thematic_on("black", "white")
plot(1:10)
plot(1:10, col = 1:10)
lattice::show.settings()

# use any hex color string
thematic_on("#444444", "#e4e4e4")
plot(1:10)
plot(1:10, col = 1:10)
lattice::show.settings()

# disables thematic (also restores global state)
thematic_off()
plot(1:10)
lattice::show.settings()

thematic_on("darkblue", "skyblue", "orange")
image(volcano)
image(volcano, col = thematic_get_option("sequential"))
lattice::show.settings()
thematic_off()

}
\seealso{
\code{\link[=sequential_gradient]{sequential_gradient()}}, \code{\link[=thematic_with_theme]{thematic_with_theme()}}, \code{\link[=thematic_save_plot]{thematic_save_plot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{font_cache_set}
\alias{font_cache_set}
\title{Control the directory used for font caching}
\usage{
font_cache_set(path, cleanup = FALSE)
}
\arguments{
\item{path}{a filepath for the new cachine directory.}

\item{cleanup}{whether or not to remove font files from the
previously used caching directory (after copying to the new location).}
}
\value{
Returns the previously used caching directory.
}
\description{
The default directory used for font caching is system dependent;
and thus, not very portable from machine to machine. Use this
function to move thematic's cache to a new path. This is primarily
useful for making font cache relative to a shiny app directory,
so that, when the app is deployed, the cache deploys with it.
}
\examples{
\dontrun{
  font_cache_set("my_app")
  shiny::runApp("my_app")
}
}
\seealso{
\code{\link[=thematic_on]{thematic_on()}}, \code{\link[=font_spec]{font_spec()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto.R
\name{auto_config}
\alias{auto_config}
\alias{auto_config_set}
\alias{auto_config_get}
\title{Configure auto theming behavior}
\usage{
auto_config(
  bg = NULL,
  fg = NULL,
  accent = NULL,
  font = NULL,
  priority = c("shiny", "config", "bslib", "rstudio")
)

auto_config_set(config)

auto_config_get()
}
\arguments{
\item{bg}{a background color.}

\item{fg}{a foreground color.}

\item{accent}{a color for making certain graphical markers 'stand out'
(e.g., the fitted line color for \code{\link[ggplot2:geom_smooth]{ggplot2::geom_smooth()}}).
Can be 2 colors for lattice (stroke vs fill accent).}

\item{font}{a \code{\link[=font_spec]{font_spec()}} object. If missing, font defaults are not altered.}

\item{priority}{the order of priority to use when resolving auto values.
Possible values include:
\itemize{
\item \code{"shiny"}: use \code{\link[shiny:getCurrentOutputInfo]{shiny::getCurrentOutputInfo()}} values (if any) to resolve auto values.
\item \code{"config"}: use the values provided to this function (if any) to resolve auto values.
\item \code{"bslib"}: use \code{bslib::bs_get_variables()} values (if any)
to resolve auto values (only relevant when knitr is in progress).
\item \code{"rstudio"}: use \code{\link[rstudioapi:getThemeInfo]{rstudioapi::getThemeInfo()}} values (if any) to resolve auto values.
}}

\item{config}{a \code{auto_config()} object.}
}
\value{
a config (list-like) object.
}
\description{
Auto theming is really only "guaranteed" to work inside of a \strong{shiny}
runtime. In any other context, auto theming is based on a set of heuristics,
which won't fit every use case. As a workaround, this function allows one
to configure both a preference for specific auto values (e.g., \code{bg}, \code{fg}, etc)
as well as the \code{priority} that certain information should receive.
}
\details{
Configuring auto theming behavior is especially useful
for developers of a custom rmarkdown output document that wish to
have more sensible auto theming behavior for users of the document.
In particular, by having the output document call \code{auto_config_set()}
"pre-knit" with the document's styling preferences (and restoring the
old defaults "post-knit"), users of the output document can then simply
call \code{thematic_on()} within their document to use those preferences.

Call this function with no arguments to get the current auto defaults.
}
\examples{
old_config <- auto_config_set(auto_config("black", "white"))
thematic_with_theme(
  thematic_theme(), {
    plot(1:10, 1:10)
 })
auto_config_set(old_config)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thematic.R
\name{okabe_ito}
\alias{okabe_ito}
\title{A color-blind safe qualitative colorscale (Okabe-Ito)}
\usage{
okabe_ito(n = NULL)
}
\arguments{
\item{n}{number of colors.}
}
\value{
a vector of color codes.
}
\description{
This is the default \code{qualitative} colorscale in \code{thematic_on()}
}
\references{
\url{https://jfly.uni-koeln.de/color/}
}
\seealso{
\code{\link[=thematic_on]{thematic_on()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto.R
\name{auto_resolve_theme}
\alias{auto_resolve_theme}
\title{Resolve auto values}
\usage{
auto_resolve_theme(theme)
}
\arguments{
\item{theme}{a \code{thematic_theme()} object.}
}
\value{
The \code{theme} object with resolved \code{'auto'} values.
}
\description{
Resolves \code{'auto'} values based on the current execution environment
and configuration (i.e., \code{\link[=auto_config_get]{auto_config_get()}}).
}
\examples{

old_config <- auto_config_set(auto_config(bg = "black", fg = "white"))

# Resolving auto values in local theme objects
theme <- thematic_theme()
theme[c("bg", "fg")]
theme <- auto_resolve_theme(theme)
theme[c("bg", "fg")]

# By default, auto values are resolved when accessing
# global theme options
thematic_on()
thematic_get_option("bg", resolve = FALSE)
thematic_get_option("bg")
thematic_off()

auto_config_set(old_config)

}
\seealso{
\code{\link[=auto_config_set]{auto_config_set()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thematic-save-plot.R
\name{thematic_save_plot}
\alias{thematic_save_plot}
\alias{default_device}
\title{Capture a thematic plot as a saved file}
\usage{
thematic_save_plot(
  expr,
  device = default_device(),
  filename = tempfile(fileext = ".png"),
  ...
)

default_device(type = c("png", "svg", "pdf", "tiff", "jpeg"))
}
\arguments{
\item{expr}{an expression that produces a plot.}

\item{device}{a graphics device to use for capturing the plot.}

\item{filename}{a filename for the produced plot. The file extension should
match the relevant \code{device}.}

\item{...}{arguments passed along to the graphics \code{device}.}

\item{type}{the type of output format}
}
\value{
\code{thematic_save_plot()} returns the \code{filename} of the produced plot
and \code{default_device()} returns a graphics device function.
}
\description{
Uses a \code{device} to capture the result of an expression (\code{expr})
that produces a plot. If \code{default_device()} is used, custom fonts
(specified through \code{\link[=font_spec]{font_spec()}}) are guaranteed to work, as long as
one of either the showtext or ragg package(s) are installed.
}
\examples{

library(thematic)
font <- font_spec("Rock Salt", scale = 1.25)
thematic_on("black", "white", font = font)
file <- thematic_save_plot(plot(1:10), res = 144)
if (interactive()) browseURL(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thematic-package.R
\docType{package}
\name{thematic-package}
\alias{thematic}
\alias{thematic-package}
\title{thematic: Unified and Automatic 'Theming' of 'ggplot2', 'lattice', and 'base' R Graphics}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Theme 'ggplot2', 'lattice', and 'base' graphics based on a few choices, including foreground color, background color, accent color, and font family. Fonts that aren't available on the system, but are available via download on 'Google Fonts', can be automatically downloaded, cached, and registered for use with the 'showtext' and 'ragg' packages.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rstudio.github.io/thematic/}
  \item \url{https://github.com/rstudio/thematic}
  \item Report bugs at \url{https://github.com/rstudio/thematic/issues}
}

}
\author{
\strong{Maintainer}: Carson Sievert \email{carson@posit.co} (\href{https://orcid.org/0000-0002-4958-2844}{ORCID})

Authors:
\itemize{
  \item Barret Schloerke \email{barret@posit.co} (\href{https://orcid.org/0000-0001-9986-114X}{ORCID})
  \item Joe Cheng \email{joe@posit.co}
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thematic.R
\name{thematic_with_theme}
\alias{thematic_with_theme}
\alias{thematic_local_theme}
\alias{thematic_set_theme}
\alias{thematic_get_theme}
\alias{thematic_get_option}
\alias{thematic_get_mixture}
\title{Tools for getting and restoring global state}
\usage{
thematic_with_theme(theme, expr)

thematic_local_theme(theme, .local_envir = parent.frame())

thematic_set_theme(theme)

thematic_get_theme(resolve = TRUE)

thematic_get_option(name = "", default = NULL, resolve = TRUE)

thematic_get_mixture(amounts = 0.5, default = NULL)
}
\arguments{
\item{theme}{a \code{thematic_theme()} object (or a return value of \link{thematic_on}/\code{\link[=thematic_get_theme]{thematic_get_theme()}})
or \code{NULL} (in which case \code{thematic_off()} is called).}

\item{expr}{R code that produces a plot.}

\item{.local_envir}{The environment to use for scoping.}

\item{resolve}{whether or not \code{'auto'} values should be resolved before returning}

\item{name}{a theme element name (e.g., \code{fg}, \code{bg}, etc.)}

\item{default}{a default value to return in the event no thematic theme is active.}

\item{amounts}{value(s) between 0 and 1 specifying how much to mix \code{bg} (0) and \code{fg} (1).}
}
\value{
the result of \code{expr}.
}
\description{
These functions are helpful for getting and/or temporarily activating a
\code{\link[=thematic_theme]{thematic_theme()}}.
}
\section{Functions}{
\itemize{
\item \code{thematic_with_theme()}: similar to \code{\link[=thematic_on]{thematic_on()}}, but for an single plot.

\item \code{thematic_local_theme()}: similar to \code{\link[=thematic_with_theme]{thematic_with_theme()}}, but de-couples
the theme from the plot expression.

\item \code{thematic_set_theme()}: set a given \code{theme} object as the current theme.

\item \code{thematic_get_theme()}: obtain the current \code{theme}.

\item \code{thematic_get_option()}: obtain a particular \code{theme} option (and provide a \code{default}
if no \code{theme} is active).

\item \code{thematic_get_mixture()}: obtain a mixture of the current \code{theme}'s \code{bg} and \code{fg}.

}}
\examples{

# Use thematic_with_theme() for a one-time use of thematic
thematic_with_theme(
  thematic_theme("darkblue", "skyblue", accent = "red"),
  plot(1:10, col = thematic_get_option("accent"), pch = 19)
)

# Use thematic_set_theme() if doing something more complicated
# like programming on top thematic (without causing side effects)
my_plot <- function(expr, las = 3, ...) {
  old_theme <- thematic_on("black", "white")
  on.exit(thematic_set_theme(old_theme), add = TRUE)
  opts <- par(las = las)
  on.exit(par(opts), add = TRUE)
  # Imagine some more customization with ...
  force(expr)
}
my_plot(plot(1:10))

thematic_off()
thematic_get_option("bg", "white")
thematic_on(bg = "red")
thematic_get_option("bg", "white")
thematic_off()

thematic_with_theme(
  thematic_theme("darkblue", "skyblue"),
  scales::show_col(thematic_get_mixture(seq(0, 1, by = 0.1)))
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thematic.R
\name{sequential_gradient}
\alias{sequential_gradient}
\title{Control parameters of the sequential colorscale}
\usage{
sequential_gradient(fg_weight = 0.9, bg_weight = 0, fg_low = TRUE, n = 30)
}
\arguments{
\item{fg_weight}{a number (between 0 and 1) defining much of the \code{fg}
color should be mixed into the colorscale.}

\item{bg_weight}{a number (between 0 and 1) defining much of the \code{bg}
color should be mixed into the colorscale.}

\item{fg_low}{if \code{TRUE} (the default), the \code{fg} color is used for the
low end of the colorscale (rather than the high end).}

\item{n}{number of color codes.}
}
\value{
a list of options for passing to the \code{sequential} argument of \code{\link[=thematic_on]{thematic_on()}}.
}
\description{
Controls the default weighting and direction of the color gradient
derived from the \code{fg}, \code{bg}, and \code{accent} color (defined in \code{thematic_on()}).
}
\examples{

# Gradient from fg to accent
fg <- sequential_gradient(1, 0)
thematic_on("black", "white", "salmon", sequential = fg)
ggplot2::qplot(1:10, 1:10, color = 1:10)

# Gradient from accent -> bg
bg <- sequential_gradient(0, 1)
thematic_on("black", "white", "salmon", sequential = bg)
ggplot2::qplot(1:10, 1:10, color = 1:10)

# Gradient from mix(accent, fg, 0.5) -> mix(accent, bg, 0.5)
mix <- sequential_gradient(0.5, 0.5)
thematic_on("black", "white", "salmon", sequential = mix)
ggplot2::qplot(1:10, 1:10, color = 1:10)

# Use fg (instead of bg) for high end of scale
mix_flip <- sequential_gradient(0.5, 0.5, fg_low = FALSE)
thematic_on("black", "white", "salmon", sequential = mix_flip)
ggplot2::qplot(1:10, 1:10, color = 1:10)

}
